from .Base import Base
from sccloud.pipeline import run_pipeline


class Clustering(Base):
    """
Run sccloud.pipeline to obtain top-level clusters.

Usage:
  sccloud cluster [options] <input_file> <output_name>
  sccloud cluster -h

Arguments:
  input_file       Input HDF5 file in 10x or sccloud format. If first-pass analysis has been performed, but you want to run some additional analysis, you could also pass a h5ad-formatted file.
  output_name      Output file name. All outputs will use it as the prefix.

Options:
  -p <number>, --threads <number>                  Number of threads. [default: 1]
  --processed                                      Input file is processed and thus no PCA & diffmap will be run.

  --genome <genome>                                A string contains comma-separated genome names. sccloud will read all groups associated with genome names in the list from the hdf5 file. If genome is None, all groups will be considered.
  --channel <channel_attr>                         Use <channel_attr> to represent different samples. This will set a 'Channel' column field with <channel_attr>.
  --black-list <black_list>                        Cell barcode attributes in black list will be poped out. Format is "attr1,attr2,...,attrn".

  --min-genes-on-raw <number>                      If input are raw 10x matrix, which include all barcodes, perform a pre-filtration step to keep the data size small. In the pre-filtration step, only keep cells with at least <number> of genes. [default: 100]
  --select-singlets                                Only select DemuxEM-predicted singlets for analysis.

  --cite-seq                                       Data are CITE-Seq data. sccloud will perform analyses on RNA count matrix first. Then it will attach the ADT matrix to the RNA matrix with all antibody names changing to 'AD-' + antibody_name. Lastly, it will embed the antibody expression using FIt-SNE (the basis used for plotting is 'citeseq_fitsne').
  --cite-seq-capping <percentile>                  For CITE-Seq surface protein expression, make all cells with expression > <percentile> to the value at <percentile> to smooth outlier. Set <percentile> to 100.0 to turn this option off. [default: 99.99]

  --output-filtration-results                      Output filtration results as a spreadsheet.
  --plot-filtration-results                        Plot filtration results as PDF files.
  --plot-filtration-figsize <figsize>              Figure size for filtration plots. <figsize> is a comma-separated list of two numbers, the width and height of the figure (e.g. 6,4).
  --output-seurat-compatible                       Output seurat-compatible h5ad file. Caution: File size might be large, do not turn this option on for large data sets.
  --output-loom                                    Output loom-formatted file.

  --min-genes <number>                             Only keep cells with at least <number> of genes. [default: 500]
  --max-genes <number>                             Only keep cells with less than <number> of genes. [default: 6000]
  --min-umis <number>                              Only keep cells with at least <number> of UMIs. [default: 100]
  --max-umis <number>                              Only keep cells with less than <number> of UMIs. [default: 600000]
  --mito-prefix <prefix>                           Prefix for mitochondrial genes. If multiple prefixes are provided, separate them by comma (e.g. "MT-,mt-"). [default: MT-]
  --percent-mito <percent>                         Only keep cells with mitochondrial percent less than <percent>%. [default: 10.0]
  --gene-percent-cells <percent>                   Only use genes that are expressed in at least <percent>% of cells to select variable genes. [default: 0.05]

  --counts-per-cell-after <number>                 Total counts per cell after normalization. [default: 1e5]

  --select-hvf-flavor <flavor>                     Highly variable feature selection method. <flavor> can be 'sccloud' or 'Seurat'. [default: sccloud]
  --select-hvf-ngenes <nfeatures>                  Select top <nfeatures> highly variable features. If <flavor> is 'Seurat' and <nfeatures> is 'None', select HVGs with z-score cutoff at 0.5. [default: 2000]
  --no-select-hvf                                  Do not select highly variable features.
  --plot-hvf                                       Plot highly variable feature selection.

  --correct-batch-effect                           Correct for batch effects.
  --batch-group-by <expression>                    Batch correction assumes the differences in gene expression between channels are due to batch effects. However, in many cases, we know that channels can be partitioned into several groups and each group is biologically different from others. In this case, we will only perform batch correction for channels within each group. This option defines the groups. If <expression> is None, we assume all channels are from one group. Otherwise, groups are defined according to <expression>. <expression> takes the form of either 'attr', or 'attr1+attr2+sccloud..+attrn', or 'attr=value11,sccloud..,value1n_1;value21,sccloud..,value2n_2;sccloud..;valuem1,sccloud..,valuemn_m'. In the first form, 'attr' should be an existing sample attribute, and groups are defined by 'attr'. In the second form, 'attr1',sccloud..,'attrn' are n existing sample attributes and groups are defined by the Cartesian product of these n attributes. In the last form, there will be m + 1 groups. A cell belongs to group i (i > 0) if and only if its sample attribute 'attr' has a value among valuei1,sccloud..,valuein_i. A cell belongs to group 0 if it does not belong to any other groups.

  --random-state <seed>                            Random number generator seed. [default: 0]
  --temp-folder <temp_folder>                      Joblib temporary folder for memmapping numpy arrays.

  --nPC <number>                                   Number of principal components. [default: 50]
  --knn-K <number>                                 Number of nearest neighbors for building kNN graph. [default: 100]
  --knn-full-speed                                 For the sake of reproducibility, we only run one thread for building kNN indices. Turn on this option will allow multiple threads to be used for index building. However, it will also reduce reproducibility due to the racing between multiple threads.

  --kBET                                           Calculate kBET.
  --kBET-batch <batch>                             kBET batch keyword.
  --kBET-alpha <alpha>                             kBET rejection alpha. [default: 0.05]
  --kBET-K <K>                                     kBET K. [default: 25]

  --diffmap                                        Calculate diffusion maps.
  --diffmap-ndc <number>                           Number of diffusion components. [default: 100]
  --diffmap-solver <solver>                        Solver for eigen decomposition, either 'eigsh' or 'randomized'. [default: eigsh]
  --diffmap-maxt <max_t>                           Maximum time stamp to search for the knee point. [default: 5000]
  --diffmap-to-3d                                  If map diffusion map into 3D space using PCA.
  --calculate-pseudotime <roots>                   Calculate diffusion-based pseudotimes based on <roots>. <roots> should be a comma-separated list of cell barcodes.

  --louvain                                        Run louvain clustering algorithm.
  --louvain-resolution <resolution>                Resolution parameter for the louvain clustering algorithm. [default: 1.3]
  --louvain-class-label <label>                    Louvain cluster label name in AnnData. [default: louvain_labels]

  --leiden                                         Run leiden clustering algorithm.
  --leiden-resolution <resolution>                 Resolution parameter for the leiden clustering algorithm. [default: 1.3]
  --leiden-niter <niter>                           Number of iterations of running the Leiden algorithm. If <niter> is negative, run Leiden iteratively until no improvement. [default: -1]
  --leiden-class-label <label>                     Leiden cluster label name in AnnData. [default: leiden_labels]

  --spectral-louvain                               Run spectral-louvain clustering algorithm.
  --spectral-louvain-basis <basis>                 Basis used for KMeans clustering. Can be 'pca' or 'diffmap'. [default: diffmap]
  --spectral-louvain-nclusters <number>            Number of clusters for Kmeans initialization. [default: 30]
  --spectral-louvain-ninit <number>                Number of Kmeans tries. [default: 20]
  --spectral-louvain-resolution <resolution>       Resolution parameter for louvain. [default: 1.3]
  --spectral-louvain-class-label <label>           Spectral-louvain label name in AnnData. [default: spectral_louvain_labels]

  --spectral-leiden                                Run spectral-leiden clustering algorithm.
  --spectral-leiden-basis <basis>                  Basis used for KMeans clustering. Can be 'pca' or 'diffmap'. [default: diffmap]
  --spectral-leiden-nclusters <number>             Number of clusters for Kmeans initialization. [default: 30]
  --spectral-leiden-ninit <number>                 Number of Kmeans tries. [default: 20]
  --spectral-leiden-resolution <resolution>        Resolution parameter for leiden. [default: 1.3]
  --spectral-leiden-class-label <label>            Spectral-leiden label name in AnnData. [default: spectral_leiden_labels]

  --tsne                                           Run multi-core t-SNE for visualization.
  --fitsne                                         Run FIt-SNE for visualization.
  --tsne-perplexity <perplexity>                   t-SNE's perplexity parameter, used by both tSNE, FItSNE and net-tSNE. [default: 30]

  --umap                                           Run umap for visualization.
  --umap-K <K>                                     K neighbors for umap. [default: 15]
  --umap-min-dist <number>                         Umap parameter. [default: 0.5]
  --umap-spread <spread>                           Umap parameter. [default: 1.0]

  --fle                                            Run force-directed layout embedding.
  --fle-K <K>                                      K neighbors for building graph for FLE. [default: 50]
  --fle-target-change-per-node <change>            Target change per node to stop forceAtlas2. [default: 2.0]
  --fle-target-steps <steps>                       Maximum number of iterations before stopping the forceAtlas2 algoritm. [default: 5000]
  --fle-memory <memory>                            Memory size in GB for the Java FA2 component. [default: 8]

  --net-down-sample-fraction <frac>                Down sampling fraction for net-related visualization. [default: 0.1]
  --net-down-sample-K <K>                          Use <K> neighbors to estimate local density for each data point for down sampling. [default: 25]
  --net-down-sample-alpha <alpha>                  Weighted down sample, proportional to radius^alpha. [default: 1.0]

  --net-regressor-L2-penalty <value>               L2 penalty parameter for the deep net regressor. [default: 0.1]

  --net-tsne                                       Run net tSNE for visualization.
  --net-tsne-polish-learning-frac <frac>           After running the deep regressor to predict new coordinates, use <frac> * nsample as the learning rate to use to polish the coordinates. [default: 0.33]
  --net-tsne-polish-niter <niter>                  Number of iterations for polishing tSNE run. [default: 150]
  --net-tsne-out-basis <basis>                     Output basis for net-tSNE. [default: net_tsne]

  --net-umap                                       Run net umap for visualization.
  --net-umap-polish-learning-rate <rate>           After running the deep regressor to predict new coordinate, what is the learning rate to use to polish the coordinates for UMAP. [default: 1.0]
  --net-umap-polish-nepochs <nepochs>              Number of iterations for polishing UMAP run. [default: 40]
  --net-umap-out-basis <basis>                     Output basis for net-UMAP. [default: net_umap]

  --net-fle                                        Run net FLE.
  --net-fle-polish-target-steps <steps>            After running the deep regressor to predict new coordinate, what is the number of force atlas 2 iterations. [default: 1500]
  --net-fle-out-basis <basis>                      Output basis for net-FLE. [default: net_fle]

  -h, --help                                       Print out help information.

Outputs:
  output_name.h5ad                 Output file in h5ad format. To load this file in python, use ``import sccloud; data = sccloud.tools.read_input('output_name.h5ad', mode = 'a')``. The log-normalized expression matrix is stored in ``data.X`` as a CSR-format sparse matrix. The ``obs`` field contains cell related attributes, including clustering results. For example, ``data.obs_names`` records cell barcodes; ``data.obs['Channel']`` records the channel each cell comes from; ``data.obs['n_genes']``, ``data.obs['n_counts']``, and ``data.obs['percent_mito']`` record the number of expressed genes, total UMI count, and mitochondrial rate for each cell respectively; ``data.obs['louvain_labels']`` and ``data.obs['approx_louvain_labels']`` record each cell's cluster labels using different clustring algorithms; ``data.obs['pseudo_time']`` records the inferred pseudotime for each cell. The ``var`` field contains gene related attributes. For example, ``data.var_names`` records gene symbols, ``data.var['gene_ids']`` records Ensembl gene IDs, and ``data.var['selected']`` records selected variable genes. The ``obsm`` field records embedding coordiates. For example, ``data.obsm['X_pca']`` records PCA coordinates, ``data.obsm['X_tsne']`` records tSNE coordinates, ``data.obsm['X_umap']`` records UMAP coordinates, ``data.obsm['X_diffmap']`` records diffusion map coordinates, ``data.obsm['X_diffmap_pca']`` records the first 3 PCs by projecting the diffusion components using PCA, and ``data.obsm['X_fle']`` records the force-directed layout coordinates from the diffusion components. The ``uns`` field stores other related information, such as reference genome (``data.uns['genome']``). If '--make-output-seurat-compatible' is on, this file can be loaded into R and converted into a Seurat object.
  output_name.seurat.h5ad          Optional output. Only exists if '--output-seurat-compatible' is set. 'output_name.h5ad' in seurat-compatible manner. This file can be loaded into R and converted into a Seurat object.
  output_name.filt.xlsx            Optional output. Only exists if '--output-filtration-results' is set. This file has two sheets --- Cell filtration stats and Gene filtration stats. The first sheet records cell filtering results and it has 10 columns: Channel, channel name; kept, number of cells kept; median_n_genes, median number of expressed genes in kept cells; median_n_umis, median number of UMIs in kept cells; median_percent_mito, median mitochondrial rate as UMIs between mitochondrial genes and all genes in kept cells; filt, number of cells filtered out; total, total number of cells before filtration, if the input contain all barcodes, this number is the cells left after '--min-genes-on-raw' filtration; median_n_genes_before, median expressed genes per cell before filtration; median_n_umis_before, median UMIs per cell before filtration; median_percent_mito_before, median mitochondrial rate per cell before filtration. The channels are sorted in ascending order with respect to the number of kept cells per channel. The second sheet records genes that failed to pass the filtering. This sheet has 3 columns: gene, gene name; n_cells, number of cells this gene is expressed; percent_cells, the fraction of cells this gene is expressed. Genes are ranked in ascending order according to number of cells the gene is expressed. Note that only genes not expressed in any cell are removed from the data. Other filtered genes are marked as non-robust and not used for TPM-like normalization.
  output_name.filt.gene.pdf        Optional output. Only exists if '--plot-filtration-results' is set. This file contains violin plots contrasting gene count distributions before and after filtration per channel.
  output_name.filt.UMI.pdf         Optional output. Only exists if '--plot-filtration-results' is set. This file contains violin plots contrasting UMI count distributions before and after filtration per channel.
  output_name.filt.mito.pdf        Optional output. Only exists if '--plot-filtration-results' is set. This file contains violin plots contrasting mitochondrial rate distributions before and after filtration per channel.
  output_name.hvf.pdf              Optional output. Only exists if '--plot-hvf' is set. This file contains a scatter plot describing the highly variable gene selection procedure.
  output_name.loom                 Optional output. Only exists if '--output-loom' is set. 'output_name.h5ad' in loom format for visualization.

Examples:
  sccloud cluster -p 20 --correct-batch-effect --louvain --tsne manton_bm_10x.h5 manton_bm
    """

    def execute(self):
        kwargs = {
            "n_jobs": int(self.args["--threads"]),
            "processed": self.args["--processed"],
            "genome": self.args["--genome"],
            "channel_attr": self.args["--channel"],
            "black_list": self.args["--black-list"],
            "subcluster": False,
            "min_genes_on_raw": int(self.args["--min-genes-on-raw"]),
            "select_singlets": self.args["--select-singlets"],
            "cite_seq": self.args["--cite-seq"],
            "cite_seq_capping": float(self.args["--cite-seq-capping"]),
            "output_filt": self.args["<output_name>"]
            if self.args["--output-filtration-results"]
            else None,
            "plot_filt": self.args["<output_name>"]
            if self.args["--plot-filtration-results"]
            else None,
            "plot_filt_figsize": self.args["--plot-filtration-figsize"],
            "seurat_compatible": self.args["--output-seurat-compatible"],
            "output_loom": self.args["--output-loom"],
            "min_genes": int(self.args["--min-genes"]),
            "max_genes": int(self.args["--max-genes"]),
            "min_umis": int(self.args["--min-umis"]),
            "max_umis": int(self.args["--max-umis"]),
            "mito_prefix": self.args["--mito-prefix"],
            "percent_mito": float(self.args["--percent-mito"]),
            "percent_cells": float(self.args["--gene-percent-cells"]),
            "norm_count": float(self.args["--counts-per-cell-after"]),
            "select_hvf": not self.args["--no-select-hvf"],
            "hvf_flavor": self.args["--select-hvf-flavor"],
            "hvf_ngenes": int(self.args["--select-hvf-ngenes"])
            if self.args["--select-hvf-ngenes"] != "None"
            else None,
            "plot_hvf": self.args["<output_name>"] if self.args["--plot-hvf"] else None,
            "batch_correction": self.args["--correct-batch-effect"],
            "group_attribute": self.args["--batch-group-by"],
            "random_state": int(self.args["--random-state"]),
            "temp_folder": self.args["--temp-folder"],
            "nPC": int(self.args["--nPC"]),
            "K": int(self.args["--knn-K"]),
            "full_speed": self.args["--knn-full-speed"],
            "kBET": self.args["--kBET"],
            "kBET_batch": self.args["--kBET-batch"],
            "kBET_alpha": float(self.args["--kBET-alpha"]),
            "kBET_K": int(self.args["--kBET-K"]),
            "diffmap": self.args["--diffmap"],
            "diffmap_ndc": int(self.args["--diffmap-ndc"]),
            "diffmap_maxt": int(self.args["--diffmap-maxt"]),
            "diffmap_solver": self.args["--diffmap-solver"],
            "diffmap_to_3d": self.args["--diffmap-to-3d"],
            "pseudotime": self.split_string(self.args["--calculate-pseudotime"]),
            "louvain": self.args["--louvain"],
            "louvain_resolution": float(self.args["--louvain-resolution"]),
            "louvain_class_label": self.args["--louvain-class-label"],
            "leiden": self.args["--leiden"],
            "leiden_resolution": float(self.args["--leiden-resolution"]),
            "leiden_niter": int(self.args["--leiden-niter"]),
            "leiden_class_label": self.args["--leiden-class-label"],
            "spectral_louvain": self.args["--spectral-louvain"],
            "spectral_louvain_basis": self.args["--spectral-louvain-basis"],
            "spectral_louvain_nclusters": int(
                self.args["--spectral-louvain-nclusters"]
            ),
            "spectral_louvain_ninit": int(self.args["--spectral-louvain-ninit"]),
            "spectral_louvain_resolution": float(
                self.args["--spectral-louvain-resolution"]
            ),
            "spectral_louvain_class_label": self.args["--spectral-louvain-class-label"],
            "spectral_leiden": self.args["--spectral-leiden"],
            "spectral_leiden_basis": self.args["--spectral-leiden-basis"],
            "spectral_leiden_nclusters": int(self.args["--spectral-leiden-nclusters"]),
            "spectral_leiden_ninit": int(self.args["--spectral-leiden-ninit"]),
            "spectral_leiden_resolution": float(
                self.args["--spectral-leiden-resolution"]
            ),
            "spectral_leiden_class_label": self.args["--spectral-leiden-class-label"],
            "tsne": self.args["--tsne"],
            "fitsne": self.args["--fitsne"],
            "tsne_perplexity": float(self.args["--tsne-perplexity"]),
            "umap": self.args["--umap"],
            "umap_K": int(self.args["--umap-K"]),
            "umap_min_dist": float(self.args["--umap-min-dist"]),
            "umap_spread": float(self.args["--umap-spread"]),
            "fle": self.args["--fle"],
            "fle_K": int(self.args["--fle-K"]),
            "fle_target_change_per_node": float(
                self.args["--fle-target-change-per-node"]
            ),
            "fle_target_steps": int(self.args["--fle-target-steps"]),
            "fle_memory": int(self.args["--fle-memory"]),
            "net_ds_frac": float(self.args["--net-down-sample-fraction"]),
            "net_ds_K": int(self.args["--net-down-sample-K"]),
            "net_ds_alpha": float(self.args["--net-down-sample-alpha"]),
            "net_l2": float(self.args["--net-regressor-L2-penalty"]),
            "net_tsne": self.args["--net-tsne"],
            "net_tsne_polish_learing_frac": float(
                self.args["--net-tsne-polish-learning-frac"]
            ),
            "net_tsne_polish_niter": int(self.args["--net-tsne-polish-niter"]),
            "net_tsne_basis": self.args["--net-tsne-out-basis"],
            "net_umap": self.args["--net-umap"],
            "net_umap_polish_learing_rate": float(
                self.args["--net-umap-polish-learning-rate"]
            ),
            "net_umap_polish_nepochs": int(self.args["--net-umap-polish-nepochs"]),
            "net_umap_basis": self.args["--net-umap-out-basis"],
            "net_fle": self.args["--net-fle"],
            "net_fle_polish_target_steps": int(
                self.args["--net-fle-polish-target-steps"]
            ),
            "net_fle_basis": self.args["--net-fle-out-basis"],
        }

        run_pipeline(self.args["<input_file>"], self.args["<output_name>"], **kwargs)



import time
import numpy as np
from sklearn.metrics.pairwise import euclidean_distances
import igraph
from collections import deque

from typing import List
import logging

logger = logging.getLogger("sccloud")


def calc_pseudotime(data: "AnnData", roots: List[str]) -> None:
    """Calculate Pseudotime based on Diffusion Map.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    roots: ``List[str]``
        List of cell barcodes in the data.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs["pseudotime"]``: Pseudotime result.

    Examples
    --------
    >>> scc.calc_pseudotime(adata, roots = list(adata.obs_names[0:100]))
    """
    start = time.time()

    if not isinstance(roots, list):
        roots = [roots]

    if "X_diffmap" not in data.obsm.keys():
        raise ValueError("Please run diffmap first!")

    data.uns["roots"] = roots
    mask = np.isin(data.obs_names, data.uns["roots"])
    distances = np.mean(
        euclidean_distances(data.obsm["X_diffmap"][mask, :], data.obsm["X_diffmap"]),
        axis=0,
    )
    dmin = distances.min()
    dmax = distances.max()
    data.obs["pseudotime"] = (distances - dmin) / (dmax - dmin)

    end = time.time()
    logger.info("calc_pseudotime finished. Time spent = {:.2f}s".format(end - start))



def calc_diffmap_dis(data: "AnnData", source: str, t: int, save_to: str) -> None:
    mask = np.isin(data.obs_names, source)
    diffmap = data.obsm["X_phi"] * (data.uns["diffmap_evals"] ** t)
    dis = euclidean_distances(diffmap[mask, :], diffmap)[0,:]
    data.obs[save_to] = 1.0 - dis


def construct_knn_graph(indices, distances):
    G = igraph.Graph(directed=False)
    G.add_vertices(indices.shape[0])
    edges = []
    w = []
    for i in range(indices.shape[0]):
        for j in range(indices.shape[1]):
            edges.append((i, indices[i][j]))
            w.append(distances[i][j])
    G.add_edges(edges)
    G.es["weight"] = w
    return G


def bfs_on_mst(G, root_id):
    mst = G.spanning_tree(weights="weight")
    myiter = mst.bfsiter(root_id, advanced=True)
    n = G.vcount()
    parents = np.full(n, -1, dtype=int)
    for value in myiter:
        if value[2] is not None:
            parents[value[0].index] = value[2].index
    return parents


def infer_path(data: "AnnData", cluster: str, clust_id, path_name: str, k: int = 10):
    """Inference on path of a cluster.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    cluster: ``str``
        Cluster name. Must exist in ``data.obs``.

    clust_id
        Cluster label. Must be a value of ``data.obs[cluster]``.

    path_name: ``str``
        Key name of the resulting path information.

    k: ``int``, optional, default: ``10``
        Number of nearest neighbors on Diffusion Map coordinates used in path reference.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs[path_name]``: The inferred path information on Diffusion Map about a specific cluster.

    Examples
    --------
    >>> scc.infer_path(adata, cluster = 'leiden_labels', clust_id = '1', path_name = 'leiden_1_path')
    """
    assert "roots" in data.uns and len(data.uns["roots"]) == 1
    root_id = int(np.isin(data.obs_names, data.uns["roots"][0]).nonzero()[0][0])
    indices = data.uns["diffmap_knn_indices"]
    distances = data.uns["diffmap_knn_distances"]
    G = construct_knn_graph(indices, distances)
    parents = bfs_on_mst(G, root_id)
    inpath = np.zeros(data.shape[0], dtype=bool)
    idx = np.isin(data.obs[cluster], clust_id)
    inpath[idx] = True

    qsize = idx.sum()
    queue = deque(idx.nonzero()[0])
    while qsize > 0:
        vid = queue.popleft()
        qsize -= 1
        if parents[vid] >= 0 and not inpath[parents[vid]]:
            inpath[parents[vid]] = True
            queue.append(parents[vid])
            qsize += 1

    for vid in np.nonzero(inpath & ~idx)[0]:
        inpath[indices[vid, 0:k]] = True

    data.obs[path_name] = inpath.astype(str)



from .Base import Base
from sccloud.pipeline import run_pipeline


class SubClustering(Base):
    """
Run sccloud to obtain subclusters.

Usage:
  sccloud subcluster [options] --subset-selection <subset-selection>... <input_file> <output_name>
  sccloud subcluster -h

Arguments:
  input_file             Single cell data with clustering done in h5ad format.
  output_name            Output file name. All outputs will use it as the prefix.

Options:
  --subset-selection <subset-selection>...         Specify which cells will be included in the subcluster analysis. Each <subset_selection> string takes the format of 'attr:value,...,value', which means select cells with attr in the values. If multiple <subset_selection> strings are specified, the subset of cells selected is the intersection of these strings.

  -p <number>, --threads <number>                  Number of threads. [default: 1]
  --correct-batch-effect                           Correct for batch effects for subclustering task.
  --batch-group-by                                 Batch correction assumes the differences in gene expression between channels are due to batch effects. However, in many cases, we know that channels can be partitioned into several groups and each group is biologically different from others. In this case, we will only perform batch correction for channels within each group. This option defines the groups. If <expression> is None, we assume all channels are from one group. Otherwise, groups are defined according to <expression>. <expression> takes the form of either 'attr', or 'attr1+attr2+sccloud..+attrn', or 'attr=value11,sccloud..,value1n_1;value21,sccloud..,value2n_2;sccloud..;valuem1,sccloud..,valuemn_m'. In the first form, 'attr' should be an existing sample attribute, and groups are defined by 'attr'. In the second form, 'attr1',sccloud..,'attrn' are n existing sample attributes and groups are defined by the Cartesian product of these n attributes. In the last form, there will be m + 1 groups. A cell belongs to group i (i > 0) if and only if its sample attribute 'attr' has a value among valuei1,sccloud..,valuein_i. A cell belongs to group 0 if it does not belong to any other groups.

  --output-loom                                    Output loom-formatted file.

  --select-hvf-flavor <flavor>                     Highly variable feature selection method. <flavor> can be 'sccloud' or 'Seurat'. [default: sccloud]
  --select-hvf-ngenes <nfeatures>                  Select top <nfeatures> highly variable features. If <flavor> is 'Seurat' and <nfeatures> is 'None', select HVGs with z-score cutoff at 0.5. [default: 2000]
  --no-select-hvf                                  Do not select highly variable features.
  --plot-hvf                                       Plot highly variable feature selection.

  --random-state <seed>                            Random number generator seed. [default: 0]
  --temp-folder <temp_folder>                      Joblib temporary folder for memmapping numpy arrays.

  --nPC <number>                                   Number of principal components. [default: 50]
  --knn-K <number>                                 Number of nearest neighbors for building kNN graph. [default: 100]
  --knn-full-speed                                 For the sake of reproducibility, we only run one thread for building kNN indices. Turn on this option will allow multiple threads to be used for index building. However, it will also reduce reproducibility due to the racing between multiple threads.

  --kBET                                           Calculate kBET.
  --kBET-batch <batch>                             kBET batch keyword.
  --kBET-alpha <alpha>                             kBET rejection alpha. [default: 0.05]
  --kBET-K <K>                                     kBET K. [default: 25]

  --diffmap                                        Calculate diffusion maps.
  --diffmap-ndc <number>                           Number of diffusion components. [default: 100]
  --diffmap-solver <solver>                        Solver for eigen decomposition, either 'eigsh' or 'randomized'. [default: eigsh]
  --diffmap-maxt <max_t>                           Maximum time stamp to search for the knee point. [default: 5000]
  --diffmap-to-3d                                  If map diffusion map into 3D space using PCA.
  --calculate-pseudotime <roots>                   Calculate diffusion-based pseudotimes based on <roots>. <roots> should be a comma-separated list of cell barcodes.

  --louvain                                        Run louvain clustering algorithm.
  --louvain-resolution <resolution>                Resolution parameter for the louvain clustering algorithm. [default: 1.3]
  --louvain-class-label <label>                    Louvain cluster label name in AnnData. [default: louvain_labels]

  --leiden                                         Run leiden clustering algorithm.
  --leiden-resolution <resolution>                 Resolution parameter for the leiden clustering algorithm. [default: 1.3]
  --leiden-niter <niter>                           Number of iterations of running the Leiden algorithm. If <niter> is negative, run Leiden iteratively until no improvement. [default: -1]
  --leiden-class-label <label>                     Leiden cluster label name in AnnData. [default: leiden_labels]

  --spectral-louvain                               Run spectral-louvain clustering algorithm.
  --spectral-louvain-basis <basis>                 Basis used for KMeans clustering. Can be 'pca' or 'diffmap'. [default: diffmap]
  --spectral-louvain-nclusters <number>            Number of clusters for Kmeans initialization. [default: 30]
  --spectral-louvain-ninit <number>                Number of Kmeans tries. [default: 20]
  --spectral-louvain-resolution <resolution>       Resolution parameter for louvain. [default: 1.3]
  --spectral-louvain-class-label <label>           Spectral-louvain label name in AnnData. [default: spectral_louvain_labels]

  --spectral-leiden                                Run spectral-leiden clustering algorithm.
  --spectral-leiden-basis <basis>                  Basis used for KMeans clustering. Can be 'pca' or 'diffmap'. [default: diffmap]
  --spectral-leiden-nclusters <number>             Number of clusters for Kmeans initialization. [default: 30]
  --spectral-leiden-ninit <number>                 Number of Kmeans tries. [default: 20]
  --spectral-leiden-resolution <resolution>        Resolution parameter for leiden. [default: 1.3]
  --spectral-leiden-class-label <label>            Spectral-leiden label name in AnnData. [default: spectral_leiden_labels]

  --tsne                                           Run multi-core t-SNE for visualization.
  --fitsne                                         Run FIt-SNE for visualization.
  --tsne-perplexity <perplexity>                   t-SNE's perplexity parameter, used by both tSNE, FItSNE net-tSNE and net-FItSNE. [default: 30]

  --umap                                           Run umap for visualization.
  --umap-K <K>                                     K neighbors for umap. [default: 15]
  --umap-min-dist <number>                         Umap parameter. [default: 0.5]
  --umap-spread <spread>                           Umap parameter. [default: 1.0]

  --fle                                            Run force-directed layout embedding.
  --fle-K <K>                                      K neighbors for building graph for FLE. [default: 50]
  --fle-target-change-per-node <change>            Target change per node to stop forceAtlas2. [default: 2.0]
  --fle-target-steps <steps>                       Maximum number of iterations before stopping the forceAtlas2 algoritm. [default: 5000]
  --fle-memory <memory>                            Memory size in GB for the Java FA2 component. [default: 8]

  --net-down-sample-fraction <frac>                Down sampling fraction for net-related visualization. [default: 0.1]
  --net-down-sample-K <K>                          Use <K> neighbors to estimate local density for each data point for down sampling. [default: 25]
  --net-down-sample-alpha <alpha>                  Weighted down sample, proportional to radius^alpha. [default: 1.0]

  --net-regressor-L2-penalty <value>               L2 penalty parameter for the deep net regressor. [default: 0.1]

  --net-tsne                                       Run net tSNE for visualization.
  --net-tsne-polish-learning-frac <frac>           After running the deep regressor to predict new coordinates, use <frac> * nsample as the learning rate to use to polish the coordinates. [default: 0.33]
  --net-tsne-polish-niter <niter>                  Number of iterations for polishing tSNE run. [default: 150]
  --net-tsne-out-basis <basis>                     Output basis for net-tSNE. [default: net_tsne]

  --net-umap                                       Run net umap for visualization.
  --net-umap-polish-learning-rate <rate>           After running the deep regressor to predict new coordinate, what is the learning rate to use to polish the coordinates for UMAP. [default: 1.0]
  --net-umap-polish-nepochs <nepochs>              Number of iterations for polishing UMAP run. [default: 40]
  --net-umap-out-basis <basis>                     Output basis for net-UMAP. [default: net_umap]

  --net-fle                                        Run net FLE.
  --net-fle-polish-target-steps <steps>            After running the deep regressor to predict new coordinate, what is the number of force atlas 2 iterations. [default: 1500]
  --net-fle-out-basis <basis>                      Output basis for net-FLE. [default: net_fle]

  -h, --help                                       Print out help information.

Outputs:
  output_name.h5ad              Output file in h5ad format. The clustering results are stored in the 'obs' field (e.g. 'louvain_labels' for louvain cluster labels). The PCA, t-SNE and diffusion map coordinates are stored in the 'obsm' field.
  output_name.loom              Optional output. Only exists if '--output-loom' is set. output_name.h5ad in loom format for visualization.

Examples:
  sccloud subcluster -p 20 --correct-batch-effect --subset-selection louvain_labels:3,6 --subset-selection Condition:CB_nonmix --tsne --louvain manton_bm.h5ad manton_bm_subset
    """

    def execute(self):
        kwargs = {
            "processed": True,
            "subcluster": True,
            "cite_seq": False,
            "select_singlets": False,
            "subset_selections": self.args["--subset-selection"],
            "n_jobs": int(self.args["--threads"]),
            "genome": None,
            "channel_attr": None,
            "black_list": None,
            "batch_correction": self.args["--correct-batch-effect"],
            "group_attribute": self.args["--batch-group-by"],
            "output_loom": self.args["--output-loom"],
            "select_hvf": not self.args["--no-select-hvf"],
            "hvf_flavor": self.args["--select-hvf-flavor"],
            "hvf_ngenes": int(self.args["--select-hvf-ngenes"])
            if self.args["--select-hvf-ngenes"] != "None"
            else None,
            "plot_hvf": self.args["<output_name>"] if self.args["--plot-hvf"] else None,
            "random_state": int(self.args["--random-state"]),
            "temp_folder": self.args["--temp-folder"],
            "nPC": int(self.args["--nPC"]),
            "K": int(self.args["--knn-K"]),
            "full_speed": self.args["--knn-full-speed"],
            "kBET": self.args["--kBET"],
            "kBET_batch": self.args["--kBET-batch"],
            "kBET_alpha": float(self.args["--kBET-alpha"]),
            "kBET_K": int(self.args["--kBET-K"]),
            "diffmap": self.args["--diffmap"],
            "diffmap_ndc": int(self.args["--diffmap-ndc"]),
            "diffmap_maxt": int(self.args["--diffmap-maxt"]),
            "diffmap_solver": self.args["--diffmap-solver"],
            "diffmap_to_3d": self.args["--diffmap-to-3d"],
            "pseudotime": self.split_string(self.args["--calculate-pseudotime"]),
            "louvain": self.args["--louvain"],
            "louvain_resolution": float(self.args["--louvain-resolution"]),
            "louvain_class_label": self.args["--louvain-class-label"],
            "leiden": self.args["--leiden"],
            "leiden_resolution": float(self.args["--leiden-resolution"]),
            "leiden_niter": int(self.args["--leiden-niter"]),
            "leiden_class_label": self.args["--leiden-class-label"],
            "spectral_louvain": self.args["--spectral-louvain"],
            "spectral_louvain_basis": self.args["--spectral-louvain-basis"],
            "spectral_louvain_nclusters": int(
                self.args["--spectral-louvain-nclusters"]
            ),
            "spectral_louvain_ninit": int(self.args["--spectral-louvain-ninit"]),
            "spectral_louvain_resolution": float(
                self.args["--spectral-louvain-resolution"]
            ),
            "spectral_louvain_class_label": self.args["--spectral-louvain-class-label"],
            "spectral_leiden": self.args["--spectral-leiden"],
            "spectral_leiden_basis": self.args["--spectral-leiden-basis"],
            "spectral_leiden_nclusters": int(self.args["--spectral-leiden-nclusters"]),
            "spectral_leiden_ninit": int(self.args["--spectral-leiden-ninit"]),
            "spectral_leiden_resolution": float(
                self.args["--spectral-leiden-resolution"]
            ),
            "spectral_leiden_class_label": self.args["--spectral-leiden-class-label"],
            "tsne": self.args["--tsne"],
            "fitsne": self.args["--fitsne"],
            "tsne_perplexity": float(self.args["--tsne-perplexity"]),
            "umap": self.args["--umap"],
            "umap_K": int(self.args["--umap-K"]),
            "umap_min_dist": float(self.args["--umap-min-dist"]),
            "umap_spread": float(self.args["--umap-spread"]),
            "fle": self.args["--fle"],
            "fle_K": int(self.args["--fle-K"]),
            "fle_target_change_per_node": float(
                self.args["--fle-target-change-per-node"]
            ),
            "fle_target_steps": int(self.args["--fle-target-steps"]),
            "fle_memory": int(self.args["--fle-memory"]),
            "net_ds_frac": float(self.args["--net-down-sample-fraction"]),
            "net_ds_K": int(self.args["--net-down-sample-K"]),
            "net_ds_alpha": float(self.args["--net-down-sample-alpha"]),
            "net_l2": float(self.args["--net-regressor-L2-penalty"]),
            "net_tsne": self.args["--net-tsne"],
            "net_tsne_polish_learing_frac": float(
                self.args["--net-tsne-polish-learning-frac"]
            ),
            "net_tsne_polish_niter": int(self.args["--net-tsne-polish-niter"]),
            "net_tsne_basis": self.args["--net-tsne-out-basis"],
            "net_umap": self.args["--net-umap"],
            "net_umap_polish_learing_rate": float(
                self.args["--net-umap-polish-learning-rate"]
            ),
            "net_umap_polish_nepochs": int(self.args["--net-umap-polish-nepochs"]),
            "net_umap_basis": self.args["--net-umap-out-basis"],
            "net_fle": self.args["--net-fle"],
            "net_fle_polish_target_steps": int(
                self.args["--net-fle-polish-target-steps"]
            ),
            "net_fle_basis": self.args["--net-fle-out-basis"],
        }

        run_pipeline(self.args["<input_file>"], self.args["<output_name>"], **kwargs)



from .Base import Base
from sccloud.tools import down_sample


class DownSample(Base):
    """
Downsample reads from a 10x molecule_info.h5 file.

Usage:
  sccloud down_sample [--random-state <seed>] <molecule_info.h5> <total_reads> <sampled_reads> <raw_feature_output_matrix_name>
  sccloud down_sample -h

Arguments:
  molecule_info.h5                    Input molecule info file (10x V3).
  total_reads                         Total number of sequenced reads.
  sampled_reads                       Target number of down-sampled reads.
  raw_feature_output_matrix_name      Output raw feature matrix name.

Options:
  --random-state <seed>        Seed for random number generator. [default: 0]
  -h, --help                   Print out help information.

Outputs:
  raw_feature_output_matrix.sccloud.h5        Output raw feature matrix.

Examples:
  sccloud down_sample --random-state 0 molecule_info.h5 400000000 100000000 output_matrix
    """

    def execute(self):
        down_sample(
            self.args["<molecule_info.h5>"],
            self.args["<raw_feature_output_matrix_name>"],
            int(self.args["<total_reads>"]),
            int(self.args["<sampled_reads>"]),
            random_state=int(self.args["--random-state"]),
        )



from .Base import Base
from sccloud.tools import aggregate_matrices


class AggregateMatrix(Base):
    """
Aggregate 10x matrices from each channel into one big matrix.

Usage:
  sccloud aggregate_matrix <csv_file> <output_name> [--restriction <restriction>... --attributes <attributes> --google-cloud --select-only-singlets --minimum-number-of-genes <ngene>]
  sccloud aggregate_matrix -h

Arguments:
  csv_file          Input csv-formatted file containing information of each scRNA-Seq run. Each row must contain at least 2 columns --- Sample, sample name and Location, location of the channel-specific count matrix in either 10x v2/v3, DGE, mtx, csv or loom format. If matrix is in DGE, mtx or csv format, an addition Reference column is required.
  output_name       The output file name.

Options:
  --restriction <restriction>...           Select channels that satisfy all restrictions. Each restriction takes the format of name:value,...,value or name:~value,..,value, where ~ refers to not. You can specifiy multiple restrictions by setting this option multiple times.
  --attributes <attributes>                Specify a comma-separated list of outputted attributes. These attributes should be column names in the csv file.
  --google-cloud                           If files are stored in google cloud. Assuming google cloud sdk is installed.
  --select-only-singlets                   If we have demultiplexed data, turning on this option will make sccloud only include barcodes that are predicted as singlets.
  --minimum-number-of-genes <ngene>        Only keep barcodes with at least <ngene> expressed genes.

  -h, --help                               Print out help information.

Outputs:
  output_name.h5sc        A sccloud-formatted HDF5 file containing the count matrices and associated attributes.

Examples:
  sccloud aggregate_matrix --restriction Source:BM,CB --restriction Individual:1-8 --attributes Source,Platform Manton_count_matrix.csv manton_bm_cb
    """

    def execute(self):
        aggregate_matrices(
            self.args["<csv_file>"],
            what_to_return=self.args["<output_name>"],
            restrictions=self.args["--restriction"],
            attributes=self.split_string(self.args["--attributes"]),
            google_cloud=self.args["--google-cloud"],
            select_singlets=self.args["--select-only-singlets"],
            ngene=self.convert_to_int(self.args["--minimum-number-of-genes"]),
        )



import time
import numpy as np
import pandas as pd
import tables
import numba
from numba import njit
from scipy.sparse import csr_matrix

from sccloud.io import write_output, Array2D, MemData


def sample_hypergeom(count, total_reads, n_sample, random_state):
    np.random.seed(random_state)
    new_count = np.zeros(count.size, dtype=int)
    for i in range(count.size):
        new_count[i] = np.random.hypergeometric(
            count[i], total_reads - count[i], n_sample
        )
        total_reads -= count[i]
        n_sample -= new_count[i]
    return new_count


@njit
def generate_sparse_matrix(barcode_idx, feature_idx):
    row_ind = []
    col_ind = []
    data = []

    start = 0
    for i in range(1, barcode_idx.size):
        if barcode_idx[start] != barcode_idx[i] or feature_idx[start] != feature_idx[i]:
            row_ind.append(barcode_idx[start])
            col_ind.append(feature_idx[start])
            data.append(i - start)
            start = i

    row_ind.append(barcode_idx[start])
    col_ind.append(feature_idx[start])
    data.append(barcode_idx.size - start)

    return row_ind, col_ind, data


def down_sample(molecule_info_file, output_name, total_reads, n_sample, random_state=0):
    with tables.open_file(molecule_info_file) as h5_in:
        barcode_idx = h5_in.get_node("/barcode_idx").read()
        feature_idx = h5_in.get_node("/feature_idx").read()
        count = h5_in.get_node("/count").read()

        new_count = sample_hypergeom(count, total_reads, n_sample, random_state)

        idx = new_count > 0
        barcode_idx = barcode_idx[idx]
        feature_idx = feature_idx[idx]

        barcodes = h5_in.get_node("/barcodes").read()
        gene_ids = h5_in.get_node("/features/id").read()
        gene_names = h5_in.get_node("/features/name").read()

        genome = h5_in.get_node("/barcode_info/genomes").read()[0].decode()

        row_ind, col_ind, data = generate_sparse_matrix(barcode_idx, feature_idx)
        mat = csr_matrix(
            (data, (row_ind, col_ind)), shape=(barcodes.size, gene_ids.size)
        )

        data = MemData()
        data.addData(
            genome,
            Array2D(
                {"barcodekey": barcodes},
                {"featurekey": gene_ids, "featurename": gene_names},
                mat,
            ),
        )

        write_output(data, output_name)

        print("Subsampled raw matrix is generated!")



import os
import unittest

import h5py

import sccloud as scc


class TestAggregate(unittest.TestCase):
    def tearDown(self):
        os.path.exists("aggregate_test.h5sc") and os.remove(
            "aggregate_test.h5sc"
        )

    def test_aggregate_10x_matrices(self):
        m1 = scc.read_input(
            "tests/scCloud-test-data/input/heart_1k_v3/filtered_feature_bc_matrix.h5",
            genome="mm10",
        )
        m2 = scc.read_input(
            "tests/scCloud-test-data/input/heart_1k_v2/filtered_gene_bc_matrices_h5.h5",
            genome="mm10",
        )
        scc.aggregate_matrices(
            "tests/scCloud-test-data/input/aggregate_test.csv",
            restrictions=[],
            attributes=["Version"],
            what_to_return="aggregate_test",
            google_cloud=False,
            select_singlets=False,
            ngene=None,
        )

        result = scc.read_input("aggregate_test.h5sc", genome="mm10")
        self.assertEqual(
            m1.shape[0] + m2.shape[0], result.shape[0], "Cell dimension is incorrect"
        )
        self.assertEqual(m1.shape[1], result.shape[1], "Gene dimension is incorrect")
        self.assertTrue(result.obs.get("Version") is not None, "Version not added")

        m1_result = result[list(range(m1.shape[0])), :]
        m2_result = result[list(range(m1.shape[0], m1.shape[0] + m2.shape[0])), :]
        self.assertEqual((m1_result.X != m1.X).sum(), 0, "Values differ")
        self.assertEqual((m2_result.X != m2.X).sum(), 0, "Values differ")
        self.assertTrue(
            m1_result.obs.index.values[0].startswith("heart_1k_v3"), "Prefix not added"
        )
        self.assertTrue(
            m2_result.obs.index.values[0].startswith("heart_1k_v2"), "Prefix not added"
        )

    def test_multi_genome(self):
        scc.aggregate_matrices(
            "tests/scCloud-test-data/input/aggregate_multi_genome.csv",
            restrictions=[],
            attributes=None,
            what_to_return="aggregate_test",
            google_cloud=False,
            select_singlets=False,
            ngene=None,
        )

        f = h5py.File("aggregate_test.h5sc", "r")
        self.assertIsNotNone(f["GRCh38"], "Genome not found")
        self.assertIsNotNone(f["mm10"], "Genome not found")
        with self.assertRaises(KeyError):
            f["mm9"]
        f.close()


if __name__ == "__main__":
    unittest.main()



from .Base import Base
from sccloud.tools import run_conversion


class PARQUET(Base):
    """
Generate a PARQUET file for web-based visualization.

Usage:
  scCloud parquet [options] <input_h5ad_file> <output_name>
  scCloud parquet -h

Arguments:
  input_h5ad_file        Analyzed single cell data in h5ad format.
  output_name            Name prefix for the parquet file.

Options:
  -p <number>, --threads <number>        Number of threads used to generate the PARQUET file. [default: 1]
  -h, --help                             Print out help information.

Outputs:
  output_name.parquet        Generated PARQUET file that contains metadata and expression levels for every gene.

Examples:
  scCloud parquet manton_bm.h5ad manton_bm
    """

    def execute(self):
        run_conversion(
            self.args["<input_h5ad_file>"],
            self.args["<output_name>"],
            nthreads=int(self.args["--threads"]),
        )



import unittest

import sccloud as scc
from .test_util import assert_adata_equal

import shutil
import os


class TestRead(unittest.TestCase):
    def tearDown(self):
        os.path.exists("test.h5ad") and os.remove("test.h5ad")
        os.path.exists("test_obsm_compound.h5ad") and os.remove(
            "test_obsm_compound.h5ad"
        )

    def test_mtx_v2(self):
        adata = scc.read_input(
            "tests/scCloud-test-data/input/hgmm_1k_filtered_gene_bc_matrices/hg19/matrix.mtx"
        )
        self.assertEqual(adata.shape[0], 504)

    def test_mtx_v3(self):
        adata = scc.read_input(
            "tests/scCloud-test-data/input/hgmm_1k_v3_filtered_feature_bc_matrix/matrix.mtx.gz"
        )
        self.assertEqual(adata.shape[0], 1046)

    def test_mtx_v2_dir(self):
        adata = scc.read_input(
            "tests/scCloud-test-data/input/hgmm_1k_filtered_gene_bc_matrices/hg19/"
        )
        self.assertEqual(adata.shape[0], 504)

    def test_mtx_v3_dir(self):
        adata = scc.read_input(
            "tests/scCloud-test-data/input/hgmm_1k_v3_filtered_feature_bc_matrix/"
        )
        self.assertEqual(adata.shape[0], 1046)

    def test_read_write_h5ad(self):
        adata = scc.read_input(
            "tests/scCloud-test-data/input/hgmm_1k_v3_filtered_feature_bc_matrix/"
        )
        scc.write_output(adata, "test.h5ad")
        adata2 = scc.read_input("test.h5ad")
        assert_adata_equal(self, adata, adata2)

    def test_read_write_old_5ad(self):
        adata = scc.read_input(
            "tests/scCloud-test-data/input/test_obsm_compound.h5ad"
        )
        scc.write_output(adata, "test.h5ad")
        adata2 = scc.read_input("test.h5ad")
        assert_adata_equal(self, adata, adata2)

    def test_read_write_old_5ad_backed(self):
        shutil.copy(
            "tests/scCloud-test-data/input/test_obsm_compound.h5ad",
            "test_obsm_compound.h5ad",
        )
        adata = scc.read_input("test_obsm_compound.h5ad", h5ad_mode="r+")
        scc.write_output(adata, "test_obsm_compound.h5ad")
        adata2 = scc.read_input("test_obsm_compound.h5ad")
        assert_adata_equal(self, adata, adata2)

    def test_read_write_old_5ad_backed_whitelist(self):
        shutil.copy(
            "tests/scCloud-test-data/input/test_obsm_compound.h5ad",
            "test_obsm_compound.h5ad",
        )
        adata = scc.read_input("test_obsm_compound.h5ad", h5ad_mode="r+")
        scc.write_output(adata, "test_obsm_compound.h5ad", whitelist=['obs'])
        adata2 = scc.read_input("test_obsm_compound.h5ad")
        assert_adata_equal(self, adata, adata2)


if __name__ == "__main__":
    unittest.main()



import os
from .Base import Base
from sccloud.annotate_cluster import run_annotate_cluster, annotate_anndata_object


class AnnotateCluster(Base):
    """
Annotate potential cell types for each cluster. This command has two forms: the first form generates putative annotations and the second form write annotations into the h5ad object.

Usage:
  sccloud annotate_cluster [--marker-file <file> --de-test <test> --de-alpha <alpha> --de-key <key> --minimum-report-score <score> --do-not-use-non-de-genes] <input_h5ad_file> <output_file>
  sccloud annotate_cluster --annotation <annotation_string> <input_h5ad_file>
  sccloud annotate_cluster -h

Arguments:
  input_h5ad_file        Single cell data with DE analysis done by sccloud de_analysis.
  output_file            Output annotation file.

Options:
  --marker-file <file>                    JSON file for markers. Could also be human_immune/mouse_immune/mouse_brain/human_brain, which triggers sccloud to markers included in the package. [default: human_immune]
  --de-test <test>                        DE test to use to infer cell types. [default: t]
  --de-alpha <alpha>                      False discovery rate to control family-wise error rate. [default: 0.05]
  --de-key <key>                          Keyword where the DE results store in varm. [default: de_res]
  --minimum-report-score <score>          Minimum cell type score to report a potential cell type. [default: 0.5]
  --do-not-use-non-de-genes               Do not count non DE genes as down-regulated.

  --annotation <annotation_string>        Write cell type annotations in <annotation_string> into <input_h5ad_file>. <annotation_string> has this format: 'anno_name:clust_name:anno_1;anno_2;...;anno_n'. 'anno_name' is the annotation attribute in the h5ad object, 'clust_name' is the attribute with cluster ids, and anno_i is the annotation for cluster i.

  -h, --help                              Print out help information.

Outputs:
  output_file        This is a text file. For each cluster, all its putative cell types are listed in descending order of the cell type score. For each putative cell type, all markers support this cell type are listed. If one putative cell type has cell subtypes, all subtypes will be listed under this cell type.

Examples:
  sccloud annotate_cluster manton_bm.h5ad manton_bm.anno.txt
  sccloud annotate_cluster --annotation "anno:T cells;B cells;NK cells;Monocytes" manton_bm.h5ad
    """

    def execute(self):
        if self.args["<output_file>"] is not None:
            run_annotate_cluster(
                self.args["<input_h5ad_file>"],
                self.args["<output_file>"],
                self.args["--marker-file"],
                de_test=self.args["--de-test"],
                de_alpha=float(self.args["--de-alpha"]),
                de_key=self.args["--de-key"],
                threshold=float(self.args["--minimum-report-score"]),
                ignore_nonde=self.args["--do-not-use-non-de-genes"],
            )
        else:
            annotate_anndata_object(
                self.args["<input_h5ad_file>"], self.args["--annotation"]
            )



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from . import estimate_background_probs, demultiplex
from sccloud.io import read_input


def down_sampling(data_gt, adt_gt, probs, n_threads=1):
    f = np.vectorize(lambda x, p: np.random.binomial(int(x + 1e-4), p, size=1)[0])

    nsample = data_gt.shape[0]
    nhto = adt_gt.X.sum()

    fracs = []
    accuracy = []
    for p in probs:
        data = data_gt.copy()
        adt = adt_gt.copy()

        adt.X.data = f(adt.X.data, p)
        idx = adt.X.sum(axis=1).A1 > 0
        adt = adt[idx,].copy()
        fracs.append(adt.X.sum() / nhto)

        estimate_background_probs(adt)
        demultiplex(data, adt, n_threads=n_threads)
        accuracy.append(
            sum(
                data.obs["assignment"].values.astype("str")
                == data_gt.obs["assignment"].values.astype("str")
            )
            / nsample
        )

    return fracs, accuracy


def plot_down_sampling(
    rna_file,
    adt_file,
    out_file,
    probs=[i / 10.0 for i in range(9, 0, -1)],
    n_threads=1,
    dpi=500,
    figsize=None,
):
    data_gt = read_input(rna_file)
    adt_gt = read_input(adt_file)
    fracs, accuracy = down_sampling(data_gt, adt_gt, probs, n_threads=n_threads)
    plt.plot(fracs, accuracy, ".-")
    ax = plt.gca()
    ax.set_xlim(1.0, 0.0)
    ax.set_ylim(0.79, 1.01)
    vals = ax.get_yticks()
    ax.set_yticklabels(["{:.0%}".format(v) for v in vals])
    ax.set_xlabel("Fraction of hashtag UMIs")
    ax.set_ylabel("Consistency")
    if figsize is not None:
        plt.gcf().set_size_inches(*figsize)
    plt.savefig(out_file, dpi=dpi)
    plt.close()



from .Base import Base
from sccloud.cite_seq import merge_rna_and_adt_data


class CITESeq(Base):
    """
Merge RNA and ADT matrices for CITE-Seq data.

Usage:
  sccloud merge_rna_adt [options] <input_raw_gene_bc_matrices_h5.h5> <input_adt_csv_file> <output_name>
  sccloud merge_rna_adt -h

Arguments:
  input_raw_gene_bc_matrices_h5.h5        Input raw RNA expression matrix in 10x hdf5 format.
  input_adt_csv_file                      Input ADT (antibody tag) count matrix in CSV format.
  output_name                             Merged output name.

Options:
  --antibody-control-csv <antibody_control_csv>         A CSV file containing the IgG control information for each antibody.

  -h, --help        Print out help information.

Outputs:
  output_name.sccloud.h5        Output file in sccloud HDF5 format. This file contains two groups --- one for RNAs, and the other for ADTs.

Examples:
  sccloud merge_rna_adt example_raw_h5.h5 example_adt.csv example_merged
  sccloud merge_rna_adt --antibody-control-csv antibody_control.csv example_raw_h5.h5 example_adt.csv example_merged
    """

    def execute(self):
        merge_rna_and_adt_data(
            self.args["<input_raw_gene_bc_matrices_h5.h5>"],
            self.args["<input_adt_csv_file>"],
            self.args["--antibody-control-csv"],
            self.args["<output_name>"],
        )



import plotly.offline as py
from natsort import natsorted

from .plot_utils import get_palettes

import numpy as np
import pandas as pd
import scipy.sparse


def scatter(df, output_file):
    labels = natsorted(df["Annotation"].unique())
    palettes = get_palettes(len(labels))
    data = []
    for i, (label, color) in enumerate(zip(labels, palettes)):
        tdf = df[df["Annotation"] == label]
        trace = dict(
            name=label,
            x=tdf.iloc[:, 1],
            y=tdf.iloc[:, 2],
            type="scattergl",
            mode="markers",
            marker=dict(size=5, color=color, line=dict(width=0)),
        )
        data.append(trace)
    layout = dict(
        title=output_file,
        xaxis=dict(title=df.columns[1]),
        yaxis=dict(title=df.columns[2]),
        margin=dict(l=0, r=0, b=0, t=0),
    )
    fig = dict(data=data, layout=layout)
    py.plot(fig, filename=output_file)


def scatter_real(df, output_file, log10=False):
    if log10:
        df["Annotation"] = np.log10(df["Annotation"])

    trace = dict(
        x=df.iloc[:, 1],
        y=df.iloc[:, 2],
        type="scattergl",
        mode="markers",
        marker=dict(
            size=5,
            color=df["Annotation"],
            colorscale="Jet",
            colorbar=dict(title="Density"),
            line=dict(width=0),
        ),
    )
    data = [trace]
    layout = dict(
        title=output_file,
        xaxis=dict(title=df.columns[1]),
        yaxis=dict(title=df.columns[2]),
        margin=dict(l=0, r=0, b=0, t=0),
    )
    fig = dict(data=data, layout=layout)
    py.plot(fig, filename=output_file)


def scatter3d(df, output_file):
    labels = natsorted(df["Annotation"].unique())
    palettes = get_palettes(len(labels))
    data = []
    for i, (label, color) in enumerate(zip(labels, palettes)):
        tdf = df[df["Annotation"] == label]
        trace = dict(
            name=label,
            x=tdf.iloc[:, 1],
            y=tdf.iloc[:, 2],
            z=tdf.iloc[:, 3],
            type="scatter3d",
            mode="markers",
            marker=dict(size=5, color=color, line=dict(width=0)),
        )
        data.append(trace)
    layout = dict(
        title=output_file,
        scene=dict(
            xaxis=dict(title=df.columns[1]),
            yaxis=dict(title=df.columns[2]),
            zaxis=dict(title=df.columns[3]),
        ),
        margin=dict(l=0, r=0, b=0, t=0),
    )
    fig = dict(data=data, layout=layout)
    py.plot(fig, filename=output_file)


def scatter3d_real(df, output_file, log10=False):
    if log10:
        df["Annotation"] = np.log10(df["Annotation"])

    trace = dict(
        x=df.iloc[:, 1],
        y=df.iloc[:, 2],
        z=df.iloc[:, 3],
        type="scatter3d",
        mode="markers",
        marker=dict(
            size=5,
            color=df["Annotation"],
            colorscale="Jet",
            colorbar=dict(title="Density"),
            line=dict(width=0),
        ),
    )
    data = [trace]
    layout = dict(
        title=output_file,
        scene=dict(
            xaxis=dict(title=df.columns[1]),
            yaxis=dict(title=df.columns[2]),
            zaxis=dict(title=df.columns[3]),
        ),
        margin=dict(l=0, r=0, b=0, t=0),
    )
    fig = dict(data=data, layout=layout)
    py.plot(fig, filename=output_file)



from .Base import Base
from sccloud.check_sample_indexes import run_check_sample_indexes


class CheckSampleIndexes(Base):
    """
Check for index collision between 10x' index sets and CITE-Seq/hashing indexes.

Usage:
  sccloud check_indexes [--num-mismatch <mismatch> --num-report <report>] <index_file>
  sccloud check_indexes -h

Arguments:
  index_file        Index file containing CITE-Seq/hashing index sequences. One sequence per line.

Options:
  --num-mismatch <mismatch>        Number of mismatch allowed for each index sequence. [default: 1]
  --num-report <report>            Number of valid 10x indexes to report. Default is to report all valid indexes. [default: 9999]
  -h, --help                       Print out help information.

Outputs:
  Up to <report> number of valid 10x indexes will be printed out to standard output.

Examples:
  sccloud check_indexes --num-report 8 index_file.txt
    """

    def execute(self):
        run_check_sample_indexes(
            self.args["<index_file>"],
            n_mis=int(self.args["--num-mismatch"]),
            n_report=int(self.args["--num-report"]),
        )



import time
import numpy as np
import pandas as pd
import logging

from scipy.sparse import issparse, csr_matrix
from scipy.stats import chi2
from sklearn.neighbors import NearestNeighbors
from joblib import effective_n_jobs
from typing import List, Tuple

from sccloud.tools import update_rep, X_from_rep, knn_is_cached

logger = logging.getLogger("sccloud")


def calculate_nearest_neighbors(
    X: np.array,
    K: int = 100,
    n_jobs: int = -1,
    method: str = "hnsw",
    M: int = 20,
    efC: int = 200,
    efS: int = 200,
    random_state: int = 0,
    full_speed: int = False,
):
    """Calculate nearest neighbors
    X is the sample by feature matrix
    Return K -1 neighbors, the first one is the point itself and thus omitted.
    TODO: Documentation
    """

    nsample = X.shape[0]

    if nsample <= 1000:
        method = "sklearn"

    if nsample < K:
        logger.warning(
            "Warning: in calculate_nearest_neighbors, number of samples = {} < K = {}!\n Set K to {}.".format(
                nsample, K, nsample
            )
        )
        K = nsample

    n_jobs = effective_n_jobs(n_jobs)

    if method == "hnsw":
        import hnswlib

        assert not issparse(X)
        # Build hnsw index
        knn_index = hnswlib.Index(space="l2", dim=X.shape[1])
        knn_index.init_index(
            max_elements=nsample, ef_construction=efC, M=M, random_seed=random_state
        )
        knn_index.set_num_threads(n_jobs if full_speed else 1)
        knn_index.add_items(X)

        # KNN query
        knn_index.set_ef(efS)
        knn_index.set_num_threads(n_jobs)
        indices, distances = knn_index.knn_query(X, k=K)
        # eliminate the first neighbor, which is the node itself
        for i in range(nsample):
            if indices[i, 0] != i:
                indices[i, 1:] = indices[i, 0:-1]
                distances[i, 1:] = distances[i, 0:-1]
        indices = indices[:, 1:].astype(int)
        distances = np.sqrt(distances[:, 1:])
    else:
        assert method == "sklearn"
        knn = NearestNeighbors(
            n_neighbors=K - 1, n_jobs=n_jobs
        )  # eliminate the first neighbor, which is the node itself
        knn.fit(X)
        distances, indices = knn.kneighbors()

    return indices, distances


def get_neighbors(
    data: "AnnData",
    K: int = 100,
    rep: str = "pca",
    n_jobs: int = -1,
    random_state: int = 0,
    full_speed: bool = False,
) -> Tuple[List[int], List[float]]:
    """Find K nearest neighbors for each data point and return the indices and distances arrays.

    Parameters
    ----------

    data : `AnnData`
        An AnnData object.
    K : `int`, optional (default: 100)
        Number of neighbors, including the data point itself.
    rep : `str`, optional (default: 'pca')
        Representation used to calculate kNN. If `None` use data.X
    n_jobs : `int`, optional (default: -1)
        Number of threads to use. -1 refers to all available threads
    random_state: `int`, optional (default: 0)
        Random seed for random number generator.
    full_speed: `bool`, optional (default: False)
        If full_speed, use multiple threads in constructing hnsw index. However, the kNN results are not reproducible. If not full_speed, use only one thread to make sure results are reproducible.

    Returns
    -------

    kNN indices and distances arrays.

    Examples
    --------
    >>> indices, distances = tools.get_neighbors(adata)
    """

    rep = update_rep(rep)
    indices_key = rep + "_knn_indices"
    distances_key = rep + "_knn_distances"

    if knn_is_cached(data, indices_key, distances_key, K):
        indices = data.uns[indices_key]
        distances = data.uns[distances_key]
        logger.info("Found cached kNN results, no calculation is required.")
    else:
        indices, distances = calculate_nearest_neighbors(
            X_from_rep(data, rep),
            K=K,
            n_jobs=effective_n_jobs(n_jobs),
            random_state=random_state,
            full_speed=full_speed,
        )
        data.uns[indices_key] = indices
        data.uns[distances_key] = distances

    return indices, distances


def get_symmetric_matrix(csr_mat: "csr_matrix") -> "csr_matrix":
    tp_mat = csr_mat.transpose().tocsr()
    sym_mat = csr_mat + tp_mat
    sym_mat.sort_indices()

    idx_mat = (csr_mat != 0).astype(int) + (tp_mat != 0).astype(int)
    idx_mat.sort_indices()
    idx = idx_mat.data == 2

    sym_mat.data[idx] /= 2.0
    return sym_mat


# We should not modify distances array!
def calculate_affinity_matrix(
    indices: List[int], distances: List[float]
) -> "csr_matrix":
    start = time.time()

    nsample = indices.shape[0]
    K = indices.shape[1]
    # calculate sigma, important to use median here!
    sigmas = np.median(distances, axis=1)
    sigmas_sq = np.square(sigmas)

    # calculate local-scaled kernel
    normed_dist = np.zeros((nsample, K), dtype=float)
    for i in range(nsample):
        numers = 2.0 * sigmas[i] * sigmas[indices[i, :]]
        denoms = sigmas_sq[i] + sigmas_sq[indices[i, :]]
        normed_dist[i, :] = np.sqrt(numers / denoms) * np.exp(
            -np.square(distances[i, :]) / denoms
        )

    W = csr_matrix(
        (normed_dist.ravel(), (np.repeat(range(nsample), K), indices.ravel())),
        shape=(nsample, nsample),
    )
    W = get_symmetric_matrix(W)

    # density normalization
    z = W.sum(axis=1).A1
    W = W.tocoo()
    W.data /= z[W.row]
    W.data /= z[W.col]
    W = W.tocsr()
    W.eliminate_zeros()

    end = time.time()
    logger.info(
        "Constructing affinity matrix is done. Time spent = {:.2f}s.".format(
            end - start
        )
    )

    return W


def neighbors(
    data: "AnnData",
    K: int = 100,
    rep: "str" = "pca",
    n_jobs: int = -1,
    random_state: int = 0,
    full_speed: bool = False,
) -> None:
    """Compute k nearest neighbors and affinity matrix, which will be used for diffmap and graph-based community detection algorithms.

    Parameters
    ----------

    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    K: ``int``, optional, default: ``100``
        Number of neighbors, including the data point itself.

    rep: ``str``, optional, default: ``"pca"``
        Embedding representation used to calculate kNN. If ``None``, use ``data.X``; otherwise, keyword ``'X_' + rep`` must exist in ``data.obsm``.
    
    n_jobs: ``int``, optional, default: ``-1``
        Number of threads to use. If ``-1``, use all available threads.
    
    random_state: ``int``, optional, default: ``0``
        Random seed set for reproducing results.
    
    full_speed: ``bool``, optional, default: ``False``
        * If ``True``, use multiple threads in constructing ``hnsw`` index. However, the kNN results are not reproducible. 
        * Otherwise, use only one thread to make sure results are reproducible.

    Returns
    -------
    ``None``

    Update ``data.uns``:
        * ``data.uns[rep + "_knn_indices"]``: kNN index matrix. Row i is the index list of kNN of cell i (excluding itself), sorted from nearest to farthest.
        * ``data.uns[rep + "_knn_distances"]``: kNN distance matrix. Row i is the distance list of kNN of cell i (excluding itselt), sorted from smallest to largest.
        * ``data.uns["W_" + rep]``: kNN graph of the data in terms of affinity matrix.

    Examples
    --------
    >>> tools.neighbors(adata)
    """

    # calculate kNN
    start = time.time()
    rep = update_rep(rep)
    indices, distances = get_neighbors(
        data,
        K=K,
        rep=rep,
        n_jobs=n_jobs,
        random_state=random_state,
        full_speed=full_speed,
    )
    end = time.time()
    logger.info("Nearest neighbor search is finished in {:.2f}s.".format(end - start))

    # calculate affinity matrix
    start = time.time()
    W = calculate_affinity_matrix(indices[:, 0 : K - 1], distances[:, 0 : K - 1])
    data.uns["W_" + rep] = W
    end = time.time()
    logger.info(
        "Affinity matrix calculation is finished in {:.2f}s".format(end - start)
    )


def calc_kBET_for_one_chunk(knn_indices, attr_values, ideal_dist, K):
    dof = ideal_dist.size - 1

    ns = knn_indices.shape[0]
    results = np.zeros((ns, 2))
    for i in range(ns):
        observed_counts = (
            pd.Series(attr_values[knn_indices[i, :]]).value_counts(sort=False).values
        )
        expected_counts = ideal_dist * K
        stat = np.sum(
            np.divide(
                np.square(np.subtract(observed_counts, expected_counts)),
                expected_counts,
            )
        )
        p_value = 1 - chi2.cdf(stat, dof)
        results[i, 0] = stat
        results[i, 1] = p_value

    return results


def calc_kBET(
    data: "AnnData",
    attr: str,
    rep: str = "pca",
    K: int = 25,
    alpha: float = 0.05,
    n_jobs: int = -1,
    random_state: int = 0,
    temp_folder: str = None,
) -> Tuple[float, float, float]:
    """Calculate the kBET metric of the data w.r.t. a specific sample attribute and embedding.

    This kBET metric is based on paper "A test metric for assessing single-cell RNA-seq batch correction" [Büttner18]_ in Nature Methods, 2018.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    attr: ``str``
        The sample attribute to consider. Must exist in ``data.obs``.

    rep: ``str``, optional, default: ``"pca"``
        The embedding representation to be used. The key ``'X_' + rep`` must exist in ``data.obsm``. By default, use PCA coordinates.

    K: ``int``, optional, default: ``25``
        Number of nearest neighbors, using L2 metric.

    alpha: ``float``, optional, default: ``0.05``
        Acceptance rate threshold. A cell is accepted if its kBET p-value is greater than or equal to ``alpha``.

    n_jobs: ``int``, optional, default: ``-1``
        Number of threads used. If ``-1``, use all available threads.

    random_state: ``int``, optional, default: ``0``
        Random seed set for reproducing results.

    temp_folder: ``str``, optional, default: ``None``
        Temporary folder for joblib execution.

    Returns
    -------
    stat_mean: ``float``
        Mean kBET chi-square statistic over all cells.

    pvalue_mean: ``float``
        Mean kBET p-value over all cells.

    accept_rate: ``float``
        kBET Acceptance rate of the sample.

    Examples
    --------
    >>> scc.calc_kBET(adata, attr = 'Channel')

    >>> scc.calc_kBET(adata, attr = 'Channel', rep = 'umap')
    """
    assert attr in data.obs
    if data.obs[attr].dtype.name != "category":
        data.obs[attr] = pd.Categorical(data.obs[attr])

    from joblib import Parallel, delayed

    ideal_dist = (
        data.obs[attr].value_counts(normalize=True, sort=False).values
    )  # ideal no batch effect distribution
    nsample = data.shape[0]
    nbatch = ideal_dist.size

    attr_values = data.obs[attr].values.copy()
    attr_values.categories = range(nbatch)

    indices, distances = get_neighbors(
        data, K=K, rep=rep, n_jobs=n_jobs, random_state=random_state
    )
    knn_indices = np.concatenate(
        (np.arange(nsample).reshape(-1, 1), indices[:, 0 : K - 1]), axis=1
    )  # add query as 1-nn

    # partition into chunks
    n_jobs = min(effective_n_jobs(n_jobs), nsample)
    starts = np.zeros(n_jobs + 1, dtype=int)
    quotient = nsample // n_jobs
    remainder = nsample % n_jobs
    for i in range(n_jobs):
        starts[i + 1] = starts[i] + quotient + (1 if i < remainder else 0)

    kBET_arr = np.concatenate(
        Parallel(n_jobs=n_jobs, max_nbytes=1e7, temp_folder=temp_folder)(
            delayed(calc_kBET_for_one_chunk)(
                knn_indices[starts[i] : starts[i + 1], :], attr_values, ideal_dist, K
            )
            for i in range(n_jobs)
        )
    )

    res = kBET_arr.mean(axis=0)
    stat_mean = res[0]
    pvalue_mean = res[1]
    accept_rate = (kBET_arr[:, 1] >= alpha).sum() / nsample

    return (stat_mean, pvalue_mean, accept_rate)


def calc_kSIM(
    data: "AnnData",
    attr: str,
    rep: str = "pca",
    K: int = 25,
    min_rate: float = 0.9,
    n_jobs: int = -1,
    random_state: int = 0,
) -> Tuple[float, float]:
    """Calculate the kSIM metric of the data w.r.t. a specific sample attribute and embedding.

    This kSIM metric measures if attr are not diffused too much.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    attr: ``str``
        The sample attribute to consider. Must exist in ``data.obs``.

    rep: ``str``, optional, default: ``"pca"``
        The embedding representation to consider. The key ``'X_' + rep`` must exist in ``data.obsm``.

    K: ``int``, optional, default: ``25``
        The number of nearest neighbors to be considered.

    min_rate: ``float``, optional, default: ``0.9``
        Acceptance rate threshold. A cell is accepted if its kSIM rate is larger than or equal to ``min_rate``.

    n_jobs: ``int``, optional, default: ``-1``
        Number of threads used. If ``-1``, use all available threads.

    random_state: ``int``, optional, default: ``0``
        Random seed set for reproducing results.

    Returns
    -------
    kSIM_mean: ``float``
        Mean kSIM rate over all the cells.

    kSIM_accept_rate: ``float``
        kSIM Acceptance rate of the sample.

    Examples
    --------
    >>> scc.calc_kSIM(adata, attr = 'cell_type')

    >>> scc.calc_kSIM(adata, attr = 'cell_type', rep = 'umap')
    """
    assert attr in data.obs
    nsample = data.shape[0]

    indices, distances = get_neighbors(
        data, K=K, rep=rep, n_jobs=n_jobs, random_state=random_state
    )
    knn_indices = np.concatenate(
        (np.arange(nsample).reshape(-1, 1), indices[:, 0 : K - 1]), axis=1
    )  # add query as 1-nn

    labels = np.reshape(data.obs[attr].values[knn_indices.ravel()], (-1, K))
    same_labs = labels == labels[:, 0].reshape(-1, 1)
    correct_rates = same_labs.sum(axis=1) / K

    kSIM_mean = correct_rates.mean()
    kSIM_accept_rate = (correct_rates >= min_rate).sum() / nsample

    return (kSIM_mean, kSIM_accept_rate)


# def calc_JSD(P, Q):
#     M = (P + Q) / 2
#     return (entropy(P, M, base = 2) + entropy(Q, M, base = 2)) / 2.0


# def calc_kBJSD_for_one_datapoint(pos, attr_values, knn_indices, ideal_dist):
#     idx = np.append(knn_indices[pos], [pos])
#     empirical_dist = pd.Series(attr_values[idx]).value_counts(normalize = True, sort = False).values
#     return calc_JSD(ideal_dist, empirical_dist)


# def calc_kBJSD(data, attr, rep = 'pca', K = 25, n_jobs = 1, random_state = 0, temp_folder = None):
#     assert attr in data.obs
#     if data.obs[attr].dtype.name != 'category':
#         data.obs[attr] = pd.Categorical(data.obs[attr])

#     from joblib import Parallel, delayed

#     ideal_dist = data.obs[attr].value_counts(normalize = True, sort = False).values # ideal no batch effect distribution
#     nsample = data.shape[0]
#     nbatch = ideal_dist.size

#     attr_values = data.obs[attr].values.copy()
#     attr_values.categories = range(nbatch)

#     indices, distances = get_neighbors(data, K = K, rep = rep, n_jobs = n_jobs, random_state = random_state)
#     knn_indices = indices[:, 0 : K - 1]
#     kBJSD_arr = np.array(Parallel(n_jobs = 1, max_nbytes = 1e7, temp_folder = temp_folder)(delayed(calc_kBJSD_for_one_datapoint)(i, attr_values, knn_indices, ideal_dist) for i in range(nsample)))

#     return kBJSD_arr.mean()



import os
from .Base import Base
from sccloud.plotting import make_static_plots


class Plotting(Base):
    """
Generate cluster composition plots.

Usage:
  sccloud plot [options] [--restriction <restriction>...] <plot_type> <input_h5ad_file> <output_file>
  sccloud plot -h

Arguments:
  plot_type              Only 2D plots, chosen from 'composition', 'scatter', 'scatter_groups', 'scatter_genes', 'scatter_gene_groups', 'heatmap', and 'qc_violin'.
  input_h5ad_file        Single cell data in h5ad file format with clustering done by 'sccloud cluster'.
  output_file            Output image file.

Options:
  --dpi <dpi>                        DPI value for the figure. [default: 500]

  --cluster-labels <attr>            Use <attr> as cluster labels. This option is used in 'composition', 'scatter_groups', 'heatmap', and 'qc_violin'.
  --attribute <attr>                 Plot <attr> against cluster labels. This option is only used in 'composition' and 'qc_violin'.
  --basis <basis>                    Basis for 2D plotting, chosen from 'tsne', 'fitsne', 'umap', 'pca', 'rpca', 'fle', 'diffmap_pca', 'net_tsne', 'net_fitsne', 'net_umap' or 'net_fle'. If CITE-Seq data is used, basis can also be 'citeseq_fitsne'. This option is used in 'scatter', 'scatter_groups', 'scatter_genes', and 'scatter_gene_groups'. [default: fitsne]
  --attributes <attrs>               <attrs> is a comma-separated list of attributes to color the basis. This option is only used in 'scatter'.
  --restriction <restriction>...     Set restriction if you only want to plot a subset of data. Multiple <restriction> strings are allowed. Each <restriction> takes the format of 'attr:value,value', or 'attr:~value,value..." which means excluding values. This option is used in 'composition' and 'scatter'.
  --apply-to-each-figure             Indicate that the <restriction> strings are not applied to all attributes but for specific attributes. The string's 'attr' value should math the attribute you want to restrict.
  --show-background                  Show points that are not selected as gray.
  --group <attr>                     <attr> is used to make group plots. In group plots, the first one contains all components in the group and the following plots show each component separately. This option is iused in 'scatter_groups' and 'scatter_gene_groups'. If <attr> is a semi-colon-separated string, parse the string as groups.
  --genes <genes>                    <genes> is a comma-separated list of gene names to visualize. This option is used in 'scatter_genes' and 'heatmap'.
  --gene <gene>                      Visualize <gene> in group plots. This option is only used in 'scatter_gene_groups'.

  --style <style>                    Composition plot styles. Can be either 'frequency', 'count', or 'normalized'. [default: frequency]
  --not-stacked                      Do not stack bars in composition plot.
  --log-y                            Plot y axis in log10 scale for composition plot.

  --nrows <nrows>                    Number of rows in the figure. If not set, sccloud will figure it out automatically.
  --ncols <ncols>                    Number of columns in the figure. If not set, sccloud will figure it out automatically.
  --subplot-size <sizes>             Sub-plot size in inches, w x h, separated by comma. Note that margins are not counted in the sizes. For composition, default is (6, 4). For scatter plots, default is (4, 4).
  --left <left>                      Figure's left margin in fraction with respect to subplot width.
  --bottom <bottom>                  Figure's bottom margin in fraction with respect to subplot height.
  --wspace <wspace>                  Horizontal space between subplots in fraction with respect to subplot width.
  --hspace <hspace>                  Vertical space between subplots in fraction with respect to subplot height.
  --alpha <alpha>                    Point transparent parameter. Can be a list of parameters separated by comma.
  --legend-fontsize <fontsize>       Legend font size.
  --use-raw                          Use anndata stored raw expression matrix. Only used by 'scatter_genes' and 'scatter_gene_groups'.

  --do-not-show-all                  Do not show all components in group for scatter_groups.

  --show-zscore                      If show zscore in heatmap.
  --heatmap-title <title>            Title for heatmap.

  --qc-type <type>                   Plot qc_violin by annotation, <type> can be either 'gene', 'count' (UMI), or 'mito' (mitochondrial rate). [default: gene]
  --qc-xtick-font <font>             X tick font for qc_violin. [default: 5]
  --qc-xtick-rotation                If rotate x label.
  --qc-line-width <width>            Line width for qc_violin. [default: 0.5]

  -h, --help                         Print out help information.

Examples:
  sccloud plot composition --cluster-labels louvain_labels --attribute Individual --style normalized --not-stacked Manton_BM.h5ad test.pdf
  sccloud plot scatter --basis tsne --attributes louvain_labels,Individual Manton_BM.h5ad test.pdf
  sccloud plot scatter_groups --cluster-labels louvain_labels --group Individual Manton_BM.h5ad test.pdf
  sccloud plot scatter_genes --genes CD8A,CD4,CD3G,MS4A1,NCAM1,CD14,ITGAX,IL3RA,CD38,CD34,PPBP Manton_BM.h5ad test.pdf
  sccloud plot scatter_gene_groups --gene CD8A --group Individual Manton_BM.h5ad test.pdf
  sccloud plot heatmap --cluster-labels louvain_labels --genes CD8A,CD4,CD3G,MS4A1,NCAM1,CD14,ITGAX,IL3RA,CD38,CD34,PPBP --heatmap-title 'markers' Manton_BM.h5ad test.pdf
  sccloud plot qc_violin --qc-type gene --cluster-labels louvain_labels --attribute Channel --subplot-size 7,5 --qc-xtick-font 5 --qc-line-width 0.5 Manton_BM.h5ad test.pdf
    """

    def execute(self):
        kwargs = {
            "cluster": self.args["--cluster-labels"],
            "attr": self.args["--attribute"],
            "restrictions": self.args["--restriction"],
            "apply_to_all": not self.args["--apply-to-each-figure"],
            "show_background": self.args["--show-background"],
            "basis": self.args["--basis"],
            "attrs": self.split_string(self.args["--attributes"]),
            "group": self.args["--group"],
            "genes": self.split_string(self.args["--genes"]),
            "gene": self.args["--gene"],
            "style": self.args["--style"],
            "stacked": not self.args["--not-stacked"],
            "logy": self.args["--log-y"],
            "nrows": int(self.args["--nrows"])
            if self.args["--nrows"] is not None
            else None,
            "ncols": int(self.args["--ncols"])
            if self.args["--ncols"] is not None
            else None,
            "subplot_size": [float(x) for x in self.args["--subplot-size"].split(",")]
            if self.args["--subplot-size"] is not None
            else None,
            "left": float(self.args["--left"])
            if self.args["--left"] is not None
            else None,
            "bottom": float(self.args["--bottom"])
            if self.args["--bottom"] is not None
            else None,
            "wspace": float(self.args["--wspace"])
            if self.args["--wspace"] is not None
            else None,
            "hspace": float(self.args["--hspace"])
            if self.args["--hspace"] is not None
            else None,
            "legend_fontsize": float(self.args["--legend-fontsize"])
            if self.args["--legend-fontsize"] is not None
            else None,
            "use_raw": self.args["--use-raw"],
            "showall": not self.args["--do-not-show-all"],
            "showzscore": self.args["--show-zscore"],
            "title": self.args["--heatmap-title"],
            "qc_type": self.args["--qc-type"],
            "qc_xtick_font": int(self.args["--qc-xtick-font"]),
            "qc_xtick_rotation": self.args["--qc-xtick-rotation"],
            "qc_line_width": float(self.args["--qc-line-width"]),
        }

        if self.args["--alpha"] is None:
            kwargs["alpha"] = None
        else:
            values = [float(x) for x in self.args["--alpha"].split(",")]
            if len(values) == 1:
                kwargs["alpha"] = values[0]
            else:
                kwargs["alpha"] = values

        make_static_plots(
            self.args["<input_h5ad_file>"],
            self.args["<plot_type>"],
            self.args["<output_file>"],
            dpi=int(self.args["--dpi"]),
            **kwargs
        )



import pandas as pd
import scipy
import numpy as np
import sccloud as sc
import subprocess
import os


def is_running_in_docker():
    return os.path.exist("/.dockerenv")


def assert_excel_equal(test_case, path, test_path):
    test_output = pd.read_excel(test_path, sheet_name=None)
    output = pd.read_excel(path, sheet_name=None)
    errors = []
    try:
        test_case.assertListEqual(list(test_output.keys()), list(output.keys()))
    except AssertionError as ex:
        errors.append(str(ex))
    for key in output:
        errors += assert_data_frames_equal(test_case, test_output[key], output[key])
    test_case.assertEqual(len(errors), 0, "\n".join(errors))


def assert_files_equal(test_case, path1, path2):
    c = subprocess.run(["diff", "-q", path1, path2])
    test_case.assertEqual(c.returncode, 0, "{} and {} differ".format(path1, path2))


def assert_dict_of_arrays_equal(test_case, dict1, dict2, blacklist):
    # TODO handle nested
    if blacklist is None:
        blacklist = set()
    keys1 = set(dict1.keys())
    keys2 = set(dict2.keys())
    errors = []
    try:
        test_case.assertSetEqual(
            keys1.difference(blacklist), keys2.difference(blacklist)
        )
    except AssertionError as ex:
        errors.append(str(ex))
    all_keys = set(list(dict1.keys()) + list(dict2.keys()))
    for key in all_keys:
        if key not in blacklist and key in keys1 and key in keys2:
            val1 = dict1[key]
            val2 = dict2[key]
            if scipy.sparse.issparse(val1):
                val1 = val1.toarray()
            if scipy.sparse.issparse(val2):
                val2 = val2.toarray()
            try:
                if isinstance(val1, np.ndarray):
                    np.testing.assert_array_almost_equal(
                        val1[~np.isnan(val1)],
                        val2[~np.isnan(val2)],
                        err_msg=str(key) + " not equal",
                    )
                else:
                    test_case.assertEqual(val1, val2, str(key) + " not equal")
                    # np.testing.assert_array_equal(val1[~np.isnan(val1)], val2[~np.isnan(val2)],
                    #     err_msg=str(key) + ' not equal')
            except AssertionError as ex:
                errors.append(str(ex))
    return errors


def assert_data_frames_equal(test_case, df1, df2, blacklist=None):
    if blacklist is None:
        blacklist = set()
    list1_keys = set(df1.columns)
    list2_keys = set(df2.columns)
    errors = []
    try:
        test_case.assertSetEqual(
            list1_keys.difference(blacklist), list2_keys.difference(blacklist)
        )
    except AssertionError as ex:
        errors.append(str(ex))
    all_keys = set(list(df1.columns) + list(df2.columns))
    for key in all_keys:
        if key not in blacklist and key in df1 and key in df2:
            val1 = df1[~df1[key].isna()][key].values
            val2 = df2[~df2[key].isna()][key].values

            try:
                if val1.dtype.kind != "O":
                    np.testing.assert_array_almost_equal(
                        val1[~np.isnan(val1)],
                        val2[~np.isnan(val2)],
                        err_msg=str(key) + " not equal",
                    )
                else:
                    np.testing.assert_array_equal(
                        val1, val2, err_msg=str(key) + " not equal"
                    )
            except AssertionError as ex:
                errors.append(str(ex))
    return errors


def assert_adata_files_equal(
    test_case,
    path,
    test_path,
    obs_blacklist=None,
    var_blacklist=None,
    uns_blacklist=None,
    obsm_blacklist=None,
    varm_blacklist=None,
):
    test_data = sc.io.read_input(test_path)
    data = sc.io.read_input(path)
    assert_adata_equal(
        test_case,
        data,
        test_data,
        obs_blacklist,
        var_blacklist,
        uns_blacklist,
        obsm_blacklist,
        varm_blacklist,
    )


def assert_adata_equal(
    test_case,
    data1,
    data2,
    obs_blacklist=None,
    var_blacklist=None,
    uns_blacklist=None,
    obsm_blacklist=None,
    varm_blacklist=None,
):
    if data1.isbacked:
        data1.X = data1.X[()]
    if data2.isbacked:
        data2.X = data2.X[()]

    if scipy.sparse.issparse(data1.X):
        data1.X = data1.X.toarray()
    if scipy.sparse.issparse(data2.X):
        data2.X = data2.X.toarray()
    errors = []
    try:
        np.testing.assert_array_almost_equal(data1.X, data2.X, err_msg="X not equal")
    except AssertionError as ex:
        errors.append(str(ex))
    if obs_blacklist is not None:
        obs1 = data1.obs.drop(obs_blacklist, axis=1)
        obs2 = data2.obs.drop(obs_blacklist, axis=1)
    else:
        obs1 = data1.obs
        obs2 = data2.obs

    errors += assert_data_frames_equal(test_case, obs1, obs2)
    if var_blacklist is not None:
        var1 = data1.var.drop(var_blacklist, axis=1)
        var2 = data2.var.drop(var_blacklist, axis=1)
    else:
        var1 = data1.var
        var2 = data2.var
    errors += assert_data_frames_equal(test_case, var1, var2)
    errors += assert_dict_of_arrays_equal(
        test_case, data1.uns, data2.uns, uns_blacklist
    )
    errors += assert_dict_of_arrays_equal(
        test_case, data1.obsm, data2.obsm, obsm_blacklist
    )
    errors += assert_dict_of_arrays_equal(
        test_case, data1.varm, data2.varm, varm_blacklist
    )
    test_case.assertEqual(len(errors), 0, "\n".join(errors))



import numpy as np
import pandas as pd
import os
import time
from subprocess import check_call

from typing import List

from sccloud.io import infer_file_format, read_input, write_output, MemData


def find_digits(value):
    pos = len(value) - 1
    while pos >= 0 and value[pos].isdigit():
        pos -= 1
    pos += 1
    assert pos < len(value)
    return (value[:pos], int(value[pos:]))


def parse_restriction_string(rstr):
    pos = rstr.index(":")
    name = rstr[:pos]
    isin = True
    if rstr[pos + 1] == "~":
        isin = False
        pos += 1
    content = set()
    for item in rstr[pos + 1 :].split(","):
        values = item.split("-")
        if len(values) == 1:
            content.add(values[0])
        else:
            prefix, fr = find_digits(values[0])
            assert values[1].isdigit()
            to = int(values[1]) + 1
            for i in range(fr, to):
                content.add(prefix + str(i))
    return (name, isin, content)


def aggregate_matrices(
    csv_file: str,
    what_to_return: str = "AnnData",
    restrictions: List[str] = [],
    attributes: List[str] = [],
    google_cloud: bool = False,
    select_singlets: bool = False,
    ngene: int = None,
    concat_matrices: bool = False,
) -> "None or AnnData or MemData":
    """Aggregate channel-specific count matrices into one big count matrix.

    This function takes as input a csv_file, which contains at least 2 columns — Sample, sample name; Location, file that contains the count matrices (e.g. filtered_gene_bc_matrices_h5.h5), and merges matrices from the same genome together. Depending on what_to_return, it can output the merged results into a sccloud-formatted HDF5 file or return as an AnnData or MemData object.

    Parameters
    ----------

    csv_file : `str`
        The CSV file containing information about each channel.
    what_to_return : `str`, optional (default: 'AnnData')
        If this value is equal to 'AnnData' or 'MemData', an AnnData or MemData object will be returned. Otherwise, results will be written into 'what_to_return.sccloud.h5' file and None is returned.
    restrictions : `list[str]`, optional (default: [])
        A list of restrictions used to select channels, each restriction takes the format of name:value,…,value or name:~value,..,value, where ~ refers to not.
    attributes : `list[str]`, optional (default: [])
        A list of attributes need to be incorporated into the output count matrix.
    google_cloud : `bool`, optional (default: False)
        If the channel-specific count matrices are stored in a google bucket.
    select_singlets : `bool`, optional (default: False)
        If we have demultiplexed data, turning on this option will make sccloud only include barcodes that are predicted as singlets.
    ngene : `int`, optional (default: None)
        The minimum number of expressed genes to keep one barcode.
    concat_matrices : `bool`, optional (default: False)
        If concatenate multiple matrices. If so, return only one AnnData object, otherwise, might return a list of AnnData objects.

    Returns
    -------

    None

    Examples
    --------
    >>> tools.aggregate_matrix('example.csv', 'example_10x.h5', ['Source:pbmc', 'Donor:1'], ['Source', 'Platform', 'Donor'])
    """

    df = pd.read_csv(csv_file, header=0, index_col="Sample")
    df["Sample"] = df.index

    # Select channels
    rvec = [parse_restriction_string(x) for x in restrictions]

    idx = pd.Series([True] * df.shape[0], index=df.index, name="selected")
    for name, isin, content in rvec:
        assert name in df.columns
        if isin:
            idx = idx & df[name].isin(content)
        else:
            idx = idx & (~(df[name].isin(content)))

    df = df.loc[idx]

    if df.shape[0] == 0:
        raise ValueError("No channels pass the restrictions!")

    # Load channels
    tot = 0
    aggrData = MemData()
    dest_paths = []
    for sample_name, row in df.iterrows():
        input_file = os.path.expanduser(
            os.path.expandvars(row["Location"].rstrip(os.sep))
        )
        file_format, copy_path, copy_type = infer_file_format(input_file)
        if google_cloud:
            base_name = os.path.basename(copy_path)
            dest_path = sample_name + "_tmp_" + base_name

            if copy_type == "directory":
                check_call(["mkdir", "-p", dest_path])
                call_args = ["gsutil", "-m", "cp", "-r", copy_path, dest_path]
            else:
                call_args = ["gsutil", "-m", "cp", copy_path, dest_path]
            check_call(call_args)
            dest_paths.append(dest_path)

            input_file = dest_path
            if file_format == "csv" and copy_type == "directory":
                input_file = os.path.join(dest_path, os.path.basename(input_file))

        genome = None
        if file_format in ["dge", "csv", "mtx", "loom"]:
            assert "Reference" in row
            genome = row["Reference"]

        data = read_input(
            input_file,
            genome=genome,
            return_type="MemData",
            ngene=ngene,
            select_singlets=select_singlets,
        )
        data.update_barcode_metadata_info(sample_name, row, attributes)
        aggrData.addAggrData(data)

        tot += 1
        print("Processed {}.".format(input_file))

    # Delete temporary file
    for dest_path in dest_paths:
        check_call(["rm", "-rf", dest_path])

    # Merge channels
    t1 = time.time()
    aggrData.aggregate()
    t2 = time.time()
    print("Data aggregation is finished in {:.2f}s.".format(t2 - t1))

    if what_to_return == "AnnData":
        aggrData = aggrData.convert_to_anndata(concat_matrices)
    elif what_to_return != "MemData":
        write_output(aggrData, what_to_return)
        aggrData = None

    print("Aggregated {tot} files.".format(tot=tot))

    return aggrData



import time
import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from sklearn.neural_network import MLPRegressor
import logging

logger = logging.getLogger("sccloud")


class MaxStdScaler(BaseEstimator, TransformerMixin):
    def __init__(self, copy=True, factor=1.0):
        self.factor = float(factor)
        self.copy = copy

    def fit(self, X):
        X = check_array(X, copy=self.copy, estimator=self, dtype=np.float64)
        self.scaler = np.max(np.std(X, axis=0, ddof=1)) / self.factor
        return self

    def transform(self, X):
        X = check_array(X, copy=self.copy, estimator=self, dtype=np.float64)
        X /= self.scaler

        return X

    def inverse_transform(self, X, copy=None):
        if copy is None:
            copy = self.copy

        X = check_array(X, copy=copy, estimator=self, dtype=np.float64)
        X *= self.scaler

        return X


def net_train_and_predict(X_train, y_train, X_pred, alpha, random_state, verbose=False):
    start_time = time.time()

    scaler_x = MaxStdScaler()
    X_train = scaler_x.fit_transform(X_train)
    scaler_y = MaxStdScaler(factor=15.0)
    y_train = scaler_y.fit_transform(y_train)

    regressor = MLPRegressor(
        hidden_layer_sizes=(100, 75, 50, 25),
        activation="relu",
        solver="sgd",
        learning_rate="adaptive",
        alpha=alpha,
        random_state=random_state,
    )
    regressor.fit(X_train, y_train)
    logger.info(regressor.loss_)

    y_pred = scaler_y.inverse_transform(
        regressor.predict(scaler_x.transform(X_pred)), copy=False
    )

    end_time = time.time()

    if verbose:
        logger.info(
            "Deep regressor traning and predicting finished. Time spent = {:.2f}s.".format(
                end_time - start_time
            )
        )

    return y_pred



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix

from sccloud.io import read_input, write_output


def merge_rna_and_adt_data(input_raw_h5, input_csv, antibody_control_csv, output_name):
    data = read_input(input_raw_h5, return_type="MemData")
    print("Loaded the RNA matrix.")

    keyword = "CITE_Seq_" + data.listKeys()[0]
    data_citeseq = read_input(input_csv, return_type="MemData", genome=keyword)
    print("Loaded the ADT matrix.")

    array2d = data_citeseq.getData(keyword)
    if antibody_control_csv is None:
        array2d.matrix = array2d.matrix.log1p()
    else:
        size = array2d.feature_metadata.shape[0]
        idx = np.zeros(size, dtype=bool)
        antibody_to_pos = pd.Series(
            data=range(size), index=array2d.feature_metadata.index
        )

        adt_matrix = array2d.matrix.toarray().astype(float)

        series = pd.read_csv(antibody_control_csv, header=0, index_col=0, squeeze=True)
        for antibody, control in series.iteritems():
            pos_a = antibody_to_pos[antibody]
            pos_c = antibody_to_pos[control]
            idx[pos_a] = True
            # convert to log expression
            adt_matrix[:, pos_a] = np.maximum(
                np.log(adt_matrix[:, pos_a] + 1.0) - np.log(adt_matrix[:, pos_c] + 1.0),
                0.0,
            )

        array2d.feature_metadata = array2d.feature_metadata[idx]
        array2d.matrix = csr_matrix(adt_matrix[:, idx])

    data.addData(keyword, array2d)
    write_output(data, output_name)

    print("Merged output is written.")


def capping(adt_matrix, percentile):
    for i in range(adt_matrix.shape[1]):
        cap = np.percentile(adt_matrix[:, i], percentile)
        adt_matrix[adt_matrix[:, i] > cap, i] = cap



import time
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, csc_matrix
from joblib import Parallel, delayed, effective_n_jobs
from statsmodels.stats.multitest import fdrcorrection as fdr
from collections import defaultdict

from typing import List, Tuple, Dict
import logging

logger = logging.getLogger("sccloud")


def calc_basic_stat(
    clust_id: str,
    data: List[float],
    indices: List[int],
    indptr: List[int],
    shape: Tuple[int, int],
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    sum_vec: List[float],
    cnt_vec: List[int],
    verbose: bool,
) -> pd.DataFrame:
    """ Calcualte basic statistics for one cluster
    """
    zero_vec = np.zeros(shape[1], dtype=np.float32)

    # recover sparse matrix
    mat = csr_matrix((data, indices, indptr), shape=shape)
    mask = cluster_labels == clust_id
    mat_clust = mat[mask]

    if cond_labels is None:
        n1 = mat_clust.shape[0]
        n2 = shape[0] - n1

        sum_clu = mat_clust.sum(axis=0).A1
        mean1 = sum_clu / n1 if n1 > 0 else zero_vec
        mean2 = (sum_vec - sum_clu) / n2 if n2 > 0 else zero_vec
        mean2[mean2 < 0.0] = 0.0

        nonzeros = mat_clust.getnnz(axis=0)
        percents = (nonzeros / n1 * 100.0).astype(np.float32) if n1 > 0 else zero_vec
        percents_other = (
            ((cnt_vec - nonzeros) / n2 * 100.0).astype(np.float32)
            if n2 > 0
            else zero_vec
        )
    else:
        cond1 = cond_labels.categories[0]
        cond_labs = cond_labels[mask]
        mask2 = cond_labs == cond1

        mat_cond1 = mat_clust[mask2]
        mat_cond2 = mat_clust[~mask2]
        n1 = mat_cond1.shape[0]
        n2 = mat_cond2.shape[0]

        mean1 = mat_cond1.mean(axis=0).A1
        mean2 = mat_cond2.mean(axis=0).A1

        percents = (
            (mat_cond1.getnnz(axis=0) / n1 * 100.0).astype(np.float32)
            if n1 > 0
            else zero_vec
        )
        percents_other = (
            (mat_cond2.getnnz(axis=0) / n2 * 100.0).astype(np.float32)
            if n2 > 0
            else zero_vec
        )

    # calculate log_fold_change and WAD, Weighted Average Difference, https://almob.biomedcentral.com/articles/10.1186/1748-7188-3-8
    log_fold_change = mean1 - mean2
    x_avg = (mean1 + mean2) / 2
    x_max = x_avg.max()
    x_min = x_avg.min() - 0.001  # to avoid divide by zero
    weights = (x_avg - x_min) / (x_max - x_min)
    WADs = log_fold_change * weights
    # calculate percent fold change
    idx = percents > 0.0
    idx_other = percents_other > 0.0
    percent_fold_change = np.zeros(shape[1], dtype=np.float32)
    percent_fold_change[(~idx) & (~idx_other)] = np.nan
    percent_fold_change[idx & (~idx_other)] = np.inf
    percent_fold_change[idx_other] = percents[idx_other] / percents_other[idx_other]

    df = pd.DataFrame(
        {
            "mean_logExpr:{0}".format(clust_id): mean1,
            "mean_logExpr_other:{0}".format(clust_id): mean2,
            "log_fold_change:{0}".format(clust_id): log_fold_change,
            "percentage:{0}".format(clust_id): percents,
            "percentage_other:{0}".format(clust_id): percents_other,
            "percentage_fold_change:{0}".format(clust_id): percent_fold_change,
            "WAD_score:{0}".format(clust_id): WADs,
        },
        index=gene_names,
    )

    if verbose:
        logger.info("calc_basic_stat finished for cluster {0}.".format(clust_id))

    return df


def collect_basic_statistics(
    X: csr_matrix,
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    n_jobs: int,
    temp_folder: str,
    verbose: bool,
) -> List[pd.DataFrame]:
    """ Collect basic statistics, triggering calc_basic_stat in parallel
    """
    start = time.time()

    sum_vec = cnt_vec = None
    if cond_labels is None:
        sum_vec = X.sum(axis=0).A1
        cnt_vec = X.getnnz(axis=0)

    result_list = Parallel(n_jobs=n_jobs, max_nbytes=1e7, temp_folder=temp_folder)(
        delayed(calc_basic_stat)(
            clust_id,
            X.data,
            X.indices,
            X.indptr,
            X.shape,
            cluster_labels,
            cond_labels,
            gene_names,
            sum_vec,
            cnt_vec,
            verbose,
        )
        for clust_id in cluster_labels.categories
    )

    end = time.time()
    if verbose:
        logger.info(
            "Collecting basic statistics is done. Time spent = {:.2f}s.".format(
                end - start
            )
        )

    return result_list


def calc_auc(
    clust_id: str,
    data: List[float],
    indices: List[int],
    indptr: List[int],
    shape: Tuple[int, int],
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    verbose: bool,
) -> pd.DataFrame:
    """ Calculate AUROC for one cluster
    """
    import sklearn.metrics as sm

    csc_mat = csc_matrix((data, indices, indptr), shape=shape)
    mask = cluster_labels == clust_id

    auroc = np.zeros(shape[1], dtype=np.float32)
    # aupr = np.zeros(shape[1], dtype = np.float32)

    if cond_labels is None:
        exprs = np.zeros(shape[0])
        y_true = mask
        n1 = mask.sum()
        n2 = shape[0] - n1
    else:
        exprs = None
        cond1 = cond_labels.categories[0]
        cond_labs = cond_labels[mask]
        y_true = cond_labs == cond1
        n1 = y_true.sum()
        n2 = mask.sum() - n1

    if n1 > 0 and n2 > 0:
        for i in range(shape[1]):
            if cond_labels is None:
                exprs[:] = 0.0
                exprs[
                    csc_mat.indices[csc_mat.indptr[i] : csc_mat.indptr[i + 1]]
                ] = csc_mat.data[csc_mat.indptr[i] : csc_mat.indptr[i + 1]]
            else:
                exprs = csc_mat[mask, i].toarray()[:, 0]

            fpr, tpr, thresholds = sm.roc_curve(y_true, exprs)
            auroc[i] = sm.auc(fpr, tpr)

            # precision, recall, thresholds = sm.precision_recall_curve(y_true, exprs)
            # aupr[i] = sm.auc(recall, precision)

    df = pd.DataFrame(
        {
            "auroc:{0}".format(clust_id): auroc,
            # "aupr:{0}".format(clust_id): aupr,
        },
        index=gene_names,
    )

    if verbose:
        logger.info("calc_auc finished for cluster {0}.".format(clust_id))

    return df


def calculate_auc_values(
    Xc: csc_matrix,
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    n_jobs: int,
    temp_folder: str,
    verbose: bool,
) -> List[pd.DataFrame]:
    """ Calculate AUROC values, triggering calc_auc in parallel
    """
    start = time.time()

    result_list = Parallel(n_jobs=n_jobs, max_nbytes=1e7, temp_folder=temp_folder)(
        delayed(calc_auc)(
            clust_id,
            Xc.data,
            Xc.indices,
            Xc.indptr,
            Xc.shape,
            cluster_labels,
            cond_labels,
            gene_names,
            verbose,
        )
        for clust_id in cluster_labels.categories
    )

    end = time.time()
    if verbose:
        logger.info(
            "AUROC values are calculated. Time spent = {:.2f}s.".format(end - start)
        )

    return result_list


def calc_t(
    clust_id: str,
    data: List[float],
    indices: List[int],
    indptr: List[int],
    shape: Tuple[int, int],
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    sum_vec: List[float],
    sum2_vec: List[float],
    verbose: bool,
) -> pd.DataFrame:
    """ Calcualte Welch's t-test for one cluster
    """

    # recover sparse matrix
    mat = csr_matrix((data, indices, indptr), shape=shape)
    pvals = np.full(shape[1], 1.0)
    mask = cluster_labels == clust_id
    mat_clust = mat[mask]

    if cond_labels is None:
        n1 = mat_clust.shape[0]
        n2 = shape[0] - n1

        if n1 > 1 and n2 > 1:
            sum_clu = mat_clust.sum(axis=0).A1
            mean1 = sum_clu / n1
            mean2 = (sum_vec - sum_clu) / n2
            mean2[mean2 < 0.0] = 0.0

            sum2_clu = mat_clust.power(2).sum(axis=0).A1
            s1sqr = (sum2_clu - n1 * (mean1 ** 2)) / (n1 - 1)
            s2sqr = ((sum2_vec - sum2_clu) - n2 * (mean2 ** 2)) / (n2 - 1)
            s2sqr[s2sqr < 0.0] = 0.0
    else:
        cond1 = cond_labels.categories[0]
        cond_labs = cond_labels[mask]
        mask2 = cond_labs == cond1

        mat_cond1 = mat_clust[mask2]
        mat_cond2 = mat_clust[~mask2]
        n1 = mat_cond1.shape[0]
        n2 = mat_cond2.shape[0]

        if n1 > 1 and n2 > 1:
            mean1 = mat_cond1.mean(axis=0).A1
            psum1 = mat_cond1.power(2).sum(axis=0).A1
            s1sqr = (psum1 - n1 * (mean1 ** 2)) / (n1 - 1)

            mean2 = mat_cond2.mean(axis=0).A1
            psum2 = mat_cond2.power(2).sum(axis=0).A1
            s2sqr = (psum2 - n2 * (mean2 ** 2)) / (n2 - 1)

    if n1 > 1 and n2 > 1:
        import scipy.stats as ss

        var_est = s1sqr / n1 + s2sqr / n2
        idx = var_est > 0.0
        if idx.sum() > 0:
            tscore = (mean1[idx] - mean2[idx]) / np.sqrt(var_est[idx])
            v = (var_est[idx] ** 2) / (
                (s1sqr[idx] / n1) ** 2 / (n1 - 1) + (s2sqr[idx] / n2) ** 2 / (n2 - 1)
            )
            pvals[idx] = ss.t.sf(np.fabs(tscore), v) * 2.0  # two-sided

    passed, qvals = fdr(pvals)

    df = pd.DataFrame(
        {
            "t_pval:{0}".format(clust_id): pvals.astype(np.float32),
            "t_qval:{0}".format(clust_id): qvals.astype(np.float32),
        },
        index=gene_names,
    )

    if verbose:
        logger.info("calc_t finished for cluster {0}.".format(clust_id))

    return df


def t_test(
    X: csr_matrix,
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    n_jobs: int,
    temp_folder: str,
    verbose: bool,
) -> List[pd.DataFrame]:
    """ Run Welch's t-test, triggering calc_t in parallel
    """
    start = time.time()

    sum_vec = sum2_vec = None
    if cond_labels is None:
        sum_vec = X.sum(axis=0).A1
        sum2_vec = X.power(2).sum(axis=0).A1

    result_list = Parallel(n_jobs=n_jobs, max_nbytes=1e7, temp_folder=temp_folder)(
        delayed(calc_t)(
            clust_id,
            X.data,
            X.indices,
            X.indptr,
            X.shape,
            cluster_labels,
            cond_labels,
            gene_names,
            sum_vec,
            sum2_vec,
            verbose,
        )
        for clust_id in cluster_labels.categories
    )

    end = time.time()
    if verbose:
        logger.info("Welch's t-test is done. Time spent = {:.2f}s.".format(end - start))

    return result_list


def calc_fisher(
    clust_id: str,
    data: List[float],
    indices: List[int],
    indptr: List[int],
    shape: Tuple[int, int],
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    cnt_vec: List[int],
    verbose: bool,
) -> pd.DataFrame:
    """ Calcualte Fisher's exact test for one cluster
    """
    import fisher

    # recover sparse matrix
    mat = csr_matrix((data, indices, indptr), shape=shape)
    mask = cluster_labels == clust_id
    mat_clust = mat[mask]

    if cond_labels is None:
        n1 = mat_clust.shape[0]
        n2 = shape[0] - n1

        a_true = mat_clust.getnnz(axis=0).astype(np.uint)
        a_false = n1 - a_true
        b_true = cnt_vec.astype(np.uint) - a_true
        b_false = n2 - b_true
    else:
        cond1 = cond_labels.categories[0]
        cond_labs = cond_labels[mask]
        mask2 = cond_labs == cond1

        mat_cond1 = mat_clust[mask2]
        mat_cond2 = mat_clust[~mask2]
        n1 = mat_cond1.shape[0]
        n2 = mat_cond2.shape[0]

        a_true = mat_cond1.getnnz(axis=0).astype(np.uint)
        a_false = n1 - a_true
        b_true = mat_cond2.getnnz(axis=0).astype(np.uint)
        b_false = n2 - b_true

    pvals = fisher.pvalue_npy(a_true, a_false, b_true, b_false)[2]
    passed, qvals = fdr(pvals)

    df = pd.DataFrame(
        {
            "fisher_pval:{0}".format(clust_id): pvals.astype(np.float32),
            "fisher_qval:{0}".format(clust_id): qvals.astype(np.float32),
        },
        index=gene_names,
    )

    if verbose:
        logger.info("calc_fisher finished for cluster {0}.".format(clust_id))

    return df


def fisher_test(
    X: csr_matrix,
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    n_jobs: int,
    temp_folder: str,
    verbose: bool,
) -> List[pd.DataFrame]:
    """ Run Fisher's exact test, triggering calc_fisher in parallel
    """
    start = time.time()

    cnt_vec = None
    if cond_labels is None:
        cnt_vec = X.getnnz(axis=0)

    result_list = Parallel(n_jobs=n_jobs, max_nbytes=1e7, temp_folder=temp_folder)(
        delayed(calc_fisher)(
            clust_id,
            X.data,
            X.indices,
            X.indptr,
            X.shape,
            cluster_labels,
            cond_labels,
            gene_names,
            cnt_vec,
            verbose,
        )
        for clust_id in cluster_labels.categories
    )

    end = time.time()
    if verbose:
        logger.info(
            "Fisher's exact test is done. Time spent = {:.2f}s.".format(end - start)
        )

    return result_list


def calc_mwu(
    clust_id: str,
    data: List[float],
    indices: List[int],
    indptr: List[int],
    shape: Tuple[int, int],
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    verbose: bool,
) -> pd.DataFrame:
    """ Run Mann-Whitney U test for one cluster
    """

    csc_mat = csc_matrix((data, indices, indptr), shape=shape)
    U_stats = np.zeros(shape[1], dtype=np.float32)
    pvals = np.full(shape[1], 1.0)
    mask = cluster_labels == clust_id

    if cond_labels is None:
        exprs = np.zeros(shape[0])
        idx_x = mask
        idx_y = ~idx_x
    else:
        exprs = None
        cond1 = cond_labels.categories[0]
        cond_labs = cond_labels[mask]
        idx_x = cond_labs == cond1
        idx_y = ~idx_x

    n1 = idx_x.sum()
    n2 = idx_y.sum()

    if n1 > 0 and n2 > 0:
        import scipy.stats as ss

        for i in range(shape[1]):
            if cond_labels is None:
                if csc_mat.indptr[i + 1] - csc_mat.indptr[i] > 0:
                    exprs[:] = 0.0
                    exprs[
                        csc_mat.indices[csc_mat.indptr[i] : csc_mat.indptr[i + 1]]
                    ] = csc_mat.data[csc_mat.indptr[i] : csc_mat.indptr[i + 1]]
                    U_stats[i], pvals[i] = ss.mannwhitneyu(
                        exprs[idx_x], exprs[idx_y], alternative="two-sided"
                    )
            else:
                tmp_mat = csc_mat[mask, i]
                if tmp_mat.data.size > 0:
                    exprs = tmp_mat.toarray()[:, 0]
                    U_stats[i], pvals[i] = ss.mannwhitneyu(
                        exprs[idx_x], exprs[idx_y], alternative="two-sided"
                    )

    passed, qvals = fdr(pvals)

    df = pd.DataFrame(
        {
            "mwu_U:{0}".format(clust_id): U_stats.astype(np.float32),
            "mwu_pval:{0}".format(clust_id): pvals.astype(np.float32),
            "mwu_qval:{0}".format(clust_id): qvals.astype(np.float32),
        },
        index=gene_names,
    )

    if verbose:
        logger.info("calc_mwu finished for cluster {0}.".format(clust_id))

    return df


def mwu_test(
    Xc: csc_matrix,
    cluster_labels: List[str],
    cond_labels: List[str],
    gene_names: List[str],
    n_jobs: int,
    temp_folder: str,
    verbose: bool,
) -> List[pd.DataFrame]:
    """ Run Mann-Whitney U test, triggering calc_mwu in parallel
    """
    start = time.time()

    result_list = Parallel(n_jobs=n_jobs, max_nbytes=1e7, temp_folder=temp_folder)(
        delayed(calc_mwu)(
            clust_id,
            Xc.data,
            Xc.indices,
            Xc.indptr,
            Xc.shape,
            cluster_labels,
            cond_labels,
            gene_names,
            verbose,
        )
        for clust_id in cluster_labels.categories
    )

    end = time.time()
    if verbose:
        logger.info(
            "Mann-Whitney U test is done. Time spent = {:.2f}s.".format(end - start)
        )

    return result_list


def organize_results(results: List[List[pd.DataFrame]]) -> pd.DataFrame:
    """ Concatenate resulting dataframes into one big dataframe
    """
    m = len(results)
    n = len(results[0])
    reslist = []

    for i in range(n):
        for j in range(m):
            reslist.append(results[j][i])

    df = pd.concat(reslist, axis=1)
    return df


def de_analysis(
    data: "AnnData",
    cluster: str,
    condition: str = None,
    subset: str = None,
    result_key: str = "de_res",
    n_jobs: int = -1,
    auc: bool = True,
    t: bool = True,
    fisher: bool = False,
    mwu: bool = False,
    temp_folder: str = None,
    verbose: bool = True,
) -> None:
    """Perform Differential Expression (DE) Analysis on data.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    cluster: ``str``
        Cluster labels used in DE analysis. Must exist in ``data.obs``.

    condition: ``str``, optional, default: ``None``
        Sample attribute used as condition in DE analysis. If ``None``, no condition is considered; otherwise, must exist in ``data.obs``.

    subset: ``str``, optional, default: ``None``
        Perform DE analysis on only a subset of cluster IDs. Cluster ID subset is specified as ``"clust_id1,clust_id2,...,clust_idn"``, where all IDs must exist in ``data.obs[cluster]``.

    result_key: ``str``, optional, default: ``"de_res"``
        Key name of DE analysis result stored.

    n_jobs: ``int``, optional, default: ``-1``
        Number of threads to use. If ``-1``, use all available threads.

    auc: ``bool``, optional, default: ``True``
        If ``True``, calculate area under ROC (AUROC) and area under Precision-Recall (AUPR).

    t: ``bool``, optional, default: ``True``
        If ``True``, calculate Welch's t test.

    fisher: ``bool``, optional, default: ``False``
        If ``True``, calculate Fisher's exact test.

    mwu: ``bool``, optional, default: ``False``
        If ``True``, calculate Mann-Whitney U test.

    temp_folder: ``str``, optional, default: ``None``
        Joblib temporary folder for memmapping numpy arrays.

    verbose: ``bool``, optional, default: ``True``
        If ``True``, show detailed intermediate output.

    Returns
    -------
    ``None``

    Update ``data.varm``:
        ``data.varm[result_key]``: DE analysis result.

    Examples
    --------
    >>> scc.de_analysis(adata, cluster = 'spectral_leiden_labels')

    subset: a comma-separated list of cluster labels. Then de will be performed only on these subsets.
    """
    start = time.time()

    if cluster not in data.obs:
        raise ValueError("Cannot find cluster label!")
    cluster_labels = data.obs[cluster].values
    if not isinstance(cluster_labels, pd.Categorical):
        cluster_labels = pd.Categorical(cluster_labels)

    cond_labels = None
    if condition is not None:
        if condition not in data.obs:
            raise ValueError("Cannot find condition!")
        cond_labels = data.obs[condition].values
        if not isinstance(cond_labels, pd.Categorical):
            cond_labels = pd.Categorical(cond_labels)
        if cond_labels.categories.size != 2:
            raise ValueError(
                "Number of distinct values in Condition is not equal to 2!"
            )

    X = data.X if isinstance(data.X, csr_matrix) else data.X[:]
    X.eliminate_zeros()  # In case there is any extra zeros

    if subset is not None:
        # subset data for de analysis
        subset = np.array(subset.split(","))
        idx_s = np.isin(subset, cluster_labels.categories.values)
        if idx_s.sum() < subset.size:
            raise ValueError(
                "These cluster labels do not exist: " + ",".join(subset[~idx_s]) + "!"
            )

        idx = np.isin(cluster_labels, subset)
        cluster_labels = cluster_labels[idx]
        if cond_labels is not None:
            cond_labels = cond_labels[idx]
        X = X[idx]

    n_jobs = effective_n_jobs(n_jobs)
    gene_names = data.var_names

    results = []
    results.append(
        collect_basic_statistics(
            X, cluster_labels, cond_labels, gene_names, n_jobs, temp_folder, verbose
        )
    )

    Xc = None
    if auc or mwu:
        t1 = time.time()
        Xc = X.tocsc()
        if verbose:
            logger.info(
                "Converting X to csc_matrix is done. Time spent = {:.2f}s.".format(
                    time.time() - t1
                )
            )

    if auc:
        results.append(
            calculate_auc_values(
                Xc,
                cluster_labels,
                cond_labels,
                gene_names,
                n_jobs,
                temp_folder,
                verbose,
            )
        )

    if t:
        results.append(
            t_test(
                X, cluster_labels, cond_labels, gene_names, n_jobs, temp_folder, verbose
            )
        )

    if fisher:
        results.append(
            fisher_test(
                X, cluster_labels, cond_labels, gene_names, n_jobs, temp_folder, verbose
            )
        )

    if mwu:
        results.append(
            mwu_test(
                Xc,
                cluster_labels,
                cond_labels,
                gene_names,
                n_jobs,
                temp_folder,
                verbose,
            )
        )

    df = organize_results(results)
    data.varm[result_key] = df.to_records(index=False)

    end = time.time()
    logger.info(
        "Differential expression analysis is finished. Time spent = {:.2f}s.".format(
            end - start
        )
    )


def get_valid_gene_index(n: int, df: pd.DataFrame, alpha: float) -> List[bool]:
    """ get genes that are DE for at least one test. If no DE tests, all genes are valid.
    """
    idx = np.zeros(n, dtype=bool)
    has_test = False
    for qval in ["t_qval", "fisher_qval", "mwu_qval"]:
        if qval in df.columns:
            idx = idx | (df[qval].values <= alpha)
            has_test = True

    if not has_test:
        idx = np.ones(n, dtype=bool)

    return idx


def get_sort_key(sort_by: List[str], col_names: List[str], direction: str):
    """ direction is either up or down; if down, do not use aupr
    """
    for key in sort_by:
        if key in col_names:
            return key
    raise ValueError("No valid key!")


def markers(
    data: "AnnData",
    head: int = None,
    de_key: str = "de_res",
    sort_by: str = "auroc,WAD_score",
    alpha: float = 0.05,
) -> Dict[str, Dict[str, pd.DataFrame]]:
    """

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.

    head: ``int``, optional, default: ``None``
        List only top ``head`` genes for each cluster. If ``None``, show any DE genes.

    de_key: ``str``, optional, default, ``de_res``
        Keyword of DE result stored in ``data.varm``.

    sort_by: ``str``, optional, default: ``"auroc,WAD_score"``
        Sort the resulting marker dictionary by ``auroc`` and ``WAD_score``.

    alpha: ``float``, optional, default: ``0.05``
        q-value threshold for getting valid DE genes. Only those with q-value of any test below ``alpha`` are significant, and thus considered as DE genes.

    Returns
    -------
    results: ``Dict[str, Dict[str, pd.DataFrame]]``
        A Python dictionary containing markers in structure ``dict[cluster_id]['up' or 'down'][dataframe]``.

    Examples
    --------
    >>> marker_dict = scc.markers(adata)
    """
    if de_key not in data.varm.keys():
        raise ValueError("Please run de_analysis first!")

    sort_by = sort_by.split(",")

    clust2cols = defaultdict(list)
    rec_array = data.varm[de_key]
    for name in rec_array.dtype.names:
        col_name, sep, clust_id = name.partition(":")
        clust2cols[clust_id].append(col_name)

    results = defaultdict(dict)
    for clust_id, col_names in clust2cols.items():
        rec_names = [x + ":" + clust_id for x in col_names]
        df = pd.DataFrame(data=rec_array[rec_names], index=data.var_names)
        df.columns = col_names
        df.index.name = "feature"

        idx = get_valid_gene_index(data.shape[1], df, alpha)

        idx_up = idx & (df["log_fold_change"].values > 0)
        df_up = df.loc[idx_up].sort_values(
            by=get_sort_key(sort_by, col_names, "up"), ascending=False, inplace=False
        )
        results[clust_id]["up"] = pd.DataFrame(
            df_up if head is None else df_up.iloc[0:head]
        )

        idx_down = idx & (df["log_fold_change"].values < 0)
        df_down = df.loc[idx_down].sort_values(
            by=get_sort_key(sort_by, col_names, "down"), ascending=True, inplace=False
        )
        results[clust_id]["down"] = pd.DataFrame(
            df_down if head is None else df_down.iloc[0:head]
        )

    return results


def write_results_to_excel(
    results: Dict[str, Dict[str, pd.DataFrame]], output_file: str, ndigits: int = 3
) -> None:
    """ Write results into Excel workbook.
    
    Parameters
    ----------
    results: ``Dict[str, Dict[str, pd.DataFrame]]``
        DE marker dictionary generated by ``scc.markers``.

    output_file: ``str``
        File name to which the marker dictionary is written.

    ndigits: ``int``, optional, default: ``3``
        Round non p-values and q-values to ``ndigits`` after decimal point in the excel.

    Returns
    -------
    ``None``

    Marker information is written to file with name ``output_file``.

    Examples
    --------
    >>> scc.write_results_to_excel(marker_dict, "result.de.xlsx")
    """
    start = time.time()

    import xlsxwriter
    from natsort import natsorted

    def format_short_output_cols(
        df_orig: pd.DataFrame, ndigits: int = 3
    ) -> pd.DataFrame:
        """ Round related float columns to ndigits decimal points.
        """
        df = pd.DataFrame(df_orig)

        cols = []
        for name in df.columns:
            if (not name.endswith("pval")) and (not name.endswith("qval")):
                cols.append(name)

        df.loc[:, cols] = df.loc[:, cols].round(ndigits)
        return df

    def add_worksheet(
        workbook: "workbook", df_orig: pd.DataFrame, sheet_name: str
    ) -> None:
        """ Add one worksheet with content as df
        """
        df = format_short_output_cols(df_orig)
        df.reset_index(inplace=True)
        worksheet = workbook.add_worksheet(name=sheet_name)

        worksheet.add_table(
            0,
            0,
            df.index.size,
            df.columns.size - 1,
            {
                "data": df.to_numpy(),
                "style": "Table Style Light 1",
                "first_column": True,
                "header_row": True,
                "columns": [{"header": x} for x in df.columns.values],
            },
        )
        # if df.shape[0] > 0:
        # else:
        #     worksheet.write_row(0, 0, df.columns.values)

    workbook = xlsxwriter.Workbook(output_file, {"nan_inf_to_errors": True})
    workbook.formats[0].set_font_size(9)

    for clust_id in natsorted(results.keys()):
        add_worksheet(workbook, results[clust_id]["up"], "up " + clust_id)
        add_worksheet(workbook, results[clust_id]["down"], "down " + clust_id)

    workbook.close()

    end = time.time()
    logger.info(
        "Excel spreadsheet is written. Time spent = {:.2f}s.".format(end - start)
    )


def run_de_analysis(
    input_file: str,
    output_excel_file: str,
    cluster: str,
    result_key: str = "de_res",
    n_jobs: int = -1,
    auc: bool = True,
    t: bool = True,
    fisher: bool = False,
    mwu: bool = False,
    temp_folder: str = None,
    verbose: bool = True,
    alpha: float = 0.05,
    ndigits: int = 3,
) -> None:
    """ For command line only
    """
    start = time.time()

    from sccloud.io import read_input, write_output

    data = read_input(input_file, h5ad_mode="r+")

    de_analysis(
        data,
        cluster,
        result_key=result_key,
        n_jobs=n_jobs,
        auc=auc,
        t=t,
        fisher=fisher,
        mwu=mwu,
        temp_folder=temp_folder,
        verbose=verbose,
    )

    write_output(data, input_file, whitelist=["varm/{}".format(result_key)])
    logger.info(
        "Differential expression results are written to varm/{} in h5ad file.".format(
            result_key
        )
    )

    results = markers(data, de_key=result_key, alpha=alpha)

    write_results_to_excel(results, output_excel_file, ndigits=ndigits)

    end = time.time()
    logger.info("run_de_analysis is finished in {:.2f}s.".format(end - start))



import numpy as np
import pandas as pd
import json
import logging

from sys import stdout
from natsort import natsorted
from typing import List, Dict, Union

logger = logging.getLogger("sccloud")


class CellType:
    def __init__(self, name: str, ignore_nonde: bool = False):
        self.name = name
        self.score = self.avgp = 0.0
        self.weak_support = []
        self.strong_support = []
        self.subtypes = None
        self.ignore_nonde = ignore_nonde

    def evaluate(
        self,
        obj: "json object",
        de_up: pd.DataFrame,
        de_down: pd.DataFrame,
        thre: float,
    ):
        """ Calculate score for matching a cluster with a putative cell type.
        """
        self.score = self.avgp = 0.0
        self.weak_support = []
        self.strong_support = []

        nump = 0
        for marker_set in obj["markers"]:
            numer = 0.0
            denom = len(marker_set["genes"]) * 2.0
            if denom == 0.0:
                continue

            for marker in marker_set["genes"]:
                sign = marker[-1]
                gsym = marker[:-1]

                if sign == "+":
                    if gsym in de_up.index:
                        fc = de_up.at[gsym, "fc"]
                        percent = de_up.at[gsym, "percent"]
                        self.avgp += percent
                        nump += 1

                        if fc >= thre:
                            numer += 2.0
                            self.strong_support.append(
                                (marker, "{0:.2f}%".format(percent))
                            )
                        else:
                            numer += 1.0 + (fc - 1.0) / (thre - 1.0)
                            self.weak_support.append(
                                (marker, "{0:.2f}%".format(percent))
                            )
                else:
                    assert sign == "-"
                    if gsym not in de_up.index:
                        if gsym in de_down.index:
                            fc = (
                                (1.0 / de_down.at[gsym, "fc"])
                                if de_down.at[gsym, "fc"] > 0.0
                                else np.inf
                            )
                            percent = de_down.at[gsym, "percent"]
                            if fc >= thre:
                                numer += 2.0
                                self.strong_support.append(
                                    (marker, "{0:.2f}%".format(percent))
                                )
                            else:
                                numer += 1.0 + (fc - 1.0) / (thre - 1.0)
                                self.weak_support.append(
                                    (marker, "{0:.2f}%".format(percent))
                                )
                        elif not self.ignore_nonde:
                            numer += 1.0
                            self.weak_support.append((marker, "N/A"))

            self.score += numer / denom * marker_set["weight"]

        self.score = (
            self.score / obj["denominator"] if obj["denominator"] > 0.0 else 0.0
        )
        if nump > 0:
            self.avgp /= nump

    def __repr__(self):
        res = "name: {0}; score: {1:.2f}; average marker percentage: {2:.2f}%".format(
            self.name, self.score, self.avgp
        )
        if len(self.strong_support) > 0:
            res += "; strong support: {0}".format(
                ",".join(["({0},{1})".format(x[0], x[1]) for x in self.strong_support])
            )
        if len(self.weak_support) > 0:
            res += "; weak support: {0}".format(
                ",".join(["({0},{1})".format(x[0], x[1]) for x in self.weak_support])
            )

        return res


class Annotator:
    def __init__(self, marker_file: Union[str, Dict], genes: List[str]) -> None:
        if type(marker_file) != dict:
            with open(marker_file) as fin:
                self.object = json.load(fin)
        else:
            self.object = marker_file
        self.recalibrate(self.object, genes)

    def recalibrate(self, obj: "json object", genes: List[str]) -> None:
        """ Remove markers that are not expressed (not in genes) and calculate partial weights for existing genes.
        """
        for celltype in obj["cell_types"]:
            denom = 0.0
            for marker_set in celltype["markers"]:
                markers = marker_set["genes"]
                s = len(markers)
                marker_set["genes"] = [x for x in markers if x[:-1] in genes]
                new_s = len(marker_set["genes"])
                marker_set["weight"] = marker_set["weight"] / s * new_s
                denom += marker_set["weight"]
            celltype["denominator"] = denom
            sub_obj = celltype.get("subtypes", None)
            if sub_obj is not None:
                self.recalibrate(sub_obj, genes)

    def evaluate(
        self,
        de_up: pd.DataFrame,
        de_down: pd.DataFrame,
        thre: float = 1.5,
        ignore_nonde: bool = False,
        obj: "json object" = None,
    ):
        """ Evaluate a cluster to determine its putative cell type.
        """
        if obj is None:
            obj = self.object

        results = []
        for celltype in obj["cell_types"]:
            ct = CellType(celltype["name"], ignore_nonde=ignore_nonde)
            ct.evaluate(celltype, de_up, de_down, thre)
            if ct.score >= 0.5:
                sub_obj = celltype.get("subtypes", None)
                if sub_obj is not None:
                    ct.subtypes = self.evaluate(
                        de_up,
                        de_down,
                        thre=thre,
                        ignore_nonde=ignore_nonde,
                        obj=sub_obj,
                    )
            results.append(ct)

        results.sort(key=lambda x: x.score, reverse=True)

        return results

    def report(
        self,
        fout: "output stream",
        ct_list: List["CellType"],
        thre: float,
        space: int = 4,
    ) -> None:
        """ Write putative cell type reports to fout.
        """
        for ct in ct_list:
            if ct.score >= thre:
                fout.write(" " * space + str(ct) + "\n")
                if ct.subtypes is not None:
                    self.report(fout, ct.subtypes, 0.5, space + 4)


def infer_cluster_names(
    cell_type_dict: Dict[str, List["CellType"]], threshold: float = 0.5
) -> List[str]:
    """Decide cluster names based on cell types automatically.

    Parameters
    ----------
    cell_type_dict: ``Dict[str, List["CellType"]]``
        Python dictionary of cell type list for each cluster. This is the output of ``scc.infer_cell_types``.

    threshold: ``float``, optional, default: ``0.5``
        A threshold for cell type result reported. It should be a real number between ``0.0`` and ``1.0``.

    Returns
    -------
    ``List[str]``
        A list of cluster names decided by their corresponding cell types. The order is consistent with clusters.

    Examples
    --------
    >>> cell_type_dict = scc.infer_cell_types(adata, markers = 'human_immune', de_test = 't')
    >>> cluster_names = scc.infer_cluster_names(cell_type_dict)
    """
    cluster_ids = natsorted(cell_type_dict.keys())
    names = []
    name_dict = dict()
    for cluster_id in cluster_ids:
        ct_list = cell_type_dict[cluster_id]

        if len(ct_list) == 0 or ct_list[0].score < threshold:
            cell_name = cluster_id
        else:
            ct = ct_list[0]
            while ct.subtypes is not None and ct.subtypes[0].score >= threshold:
                ct = ct.subtypes[0]
            cell_name = ct.name

            if cell_name in name_dict:
                name_dict[cell_name] += 1
                cell_name = cell_name + "-" + str(name_dict[cell_name])
            else:
                name_dict[cell_name] = 1

        names.append(cell_name)

    return names


def infer_cell_types(
    data: "AnnData",
    markers: Union[str, Dict],
    de_test: str,
    de_alpha: float = 0.05,
    de_key: str = "de_res",
    threshold: float = 0.5,
    ignore_nonde: bool = False,
    output_file: str = None,
) -> Dict[str, List["CellType"]]:
    """Infer putative cell types for each cluster using legacy markers.

    Parameters
    ----------

    data : ``anndata.AnnData``
        AnnData object of count matrix and DE analysis results.

    markers : ``str`` or ``Dict``
        * If ``str``, it 
            * either refers to a JSON file containing legacy markers, or
            * ``'human_immune'`` for predefined sccloud markers on human immune cells;
            * ``'mouse_immune'`` for mouse immune cells;
            * ``'human_brain'`` for human brain cells;
            * ``'mouse_brain'`` for mouse brain cells.
        * If ``Dict``, it refers to a Python dictionary describing the markers.

    de_test: ``str``
        sccloud determines cell types using DE test results. This argument indicates which DE test result to use, can be either ``'t'``, ``'fisher'`` or ``'mwu'``.

    de_alpha: ``float``, optional, default: ``0.05``
        False discovery rate for controling family-wide error.

    de_key : ``str``, optional, default: ``"de_res"``
        The keyword in ``data.varm`` that stores DE analysis results.

    threshold : ``float``, optional, defaut: ``0.5``
        Only report putative cell types with a score larger than or equal to ``threshold``.

    ignore_nonde: ``bool``, optional, default: ``False``
        Do not consider non DE genes as weak negative markers.

    output_file: ``str``, optional, default: ``None``
        File name of output cluster annotation. If ``None``, do not write to any file.

    Returns
    -------
    ``Dict[str, List["CellType"]]``
        Python dictionary with cluster ID's being keys, and their corresponding cell type lists sortec by scores being values.

    Examples
    --------
    >>> cell_type_dict = scc.infer_cell_types(adata, markers = 'human_immune', de_test = 't')
    """

    if output_file is not None:
        fout = open(output_file, "w")

    predefined_markers = dict(
        human_immune="human_immune_cell_markers.json",
        mouse_immune="mouse_immune_cell_markers.json",
        mouse_brain="mouse_brain_cell_markers.json",
        human_brain="human_brain_cell_markers.json",
    )

    if markers in predefined_markers:
        import pkg_resources

        markers = pkg_resources.resource_filename(
            "sccloud.annotate_cluster", predefined_markers[markers]
        )

    anno = Annotator(markers, data.var_names)

    clusts = natsorted(
        [
            x.rpartition(":")[2]
            for x in data.varm[de_key].dtype.names
            if x.startswith("WAD_score:")
        ]
    )
    cell_type_results = {}
    for clust_id in clusts:
        idx = data.varm[de_key]["{0}_qval:{1}".format(de_test, clust_id)] <= de_alpha

        idx_up = idx & (data.varm[de_key]["log_fold_change:{0}".format(clust_id)] > 0.0)
        idx_down = idx & (
            data.varm[de_key]["log_fold_change:{0}".format(clust_id)] < 0.0
        )
        assert idx_up.sum() + idx_down.sum() == idx.sum()

        cols = [
            "{0}:{1}".format(x, clust_id)
            for x in [
                "percentage_fold_change" if de_test == "fisher" else "log_fold_change",
                "percentage",
            ]
        ]
        de_up = pd.DataFrame(
            data=data.varm[de_key][cols][idx_up], index=data.var_names[idx_up]
        )
        de_up.rename(columns={cols[0]: "fc", cols[1]: "percent"}, inplace=True)
        de_down = pd.DataFrame(
            data=data.varm[de_key][cols][idx_down], index=data.var_names[idx_down]
        )
        de_down.rename(columns={cols[0]: "fc", cols[1]: "percent"}, inplace=True)

        if de_test != "fisher":
            de_up["fc"] = np.exp(de_up["fc"])
            de_down["fc"] = np.exp(de_down["fc"])

        results = anno.evaluate(de_up, de_down, ignore_nonde=ignore_nonde)

        if output_file is not None:
            fout.write("Cluster {}:\n".format(clust_id))
            anno.report(fout, results, threshold)

        cell_type_results[clust_id] = results

    if output_file is not None:
        fout.close()

    return cell_type_results


def annotate(
    data: "AnnData", name: str, based_on: str, anno_dict: Dict[str, str]
) -> None:
    """Add annotation to AnnData obj.

    Parameters
    ----------

    data : `AnnData`
        AnnData object.
    name : `str`
        Name of the new annotation in data.obs.
    based_on : `str`
        Name of the attribute the cluster ids coming from.
    anno_dict : `Dict[str, str]`
        Dictionary mapping from cluster id to cell type.

    Returns
    -------

    None

    Examples
    --------
    >>> annotate_cluster.annotate(data, 'anno', 'spectral_louvain_labels', {'1': 'T cell', '2': 'B cell'})
    """
    data.obs[name] = [anno_dict[x] for x in data.obs[based_on]]


def run_annotate_cluster(
    input_file: str,
    output_file: str,
    marker_file: str,
    de_test: str,
    de_alpha: float = 0.05,
    de_key: str = "de_res",
    threshold: float = 0.5,
    ignore_nonde: bool = False,
) -> None:
    """ For command line use.
    """
    import time
    from sccloud.io import read_input

    start = time.time()
    data = read_input(input_file, h5ad_mode="r")
    infer_cell_types(
        data,
        marker_file,
        de_test,
        de_alpha=de_alpha,
        de_key=de_key,
        threshold=threshold,
        ignore_nonde=ignore_nonde,
        output_file=output_file,
    )
    data.file.close()
    end = time.time()
    logger.info("Time spent for annotating clusters is {:.2f}s.".format(end - start))


def annotate_anndata_object(input_file: str, annotation: str) -> None:
    """ For command line use.
        annotation:  anno_name:clust_name:cell_type1;...cell_typen
    """
    from sccloud.io import read_input, write_output

    data = read_input(input_file, h5ad_mode="r+")
    anno_name, clust_name, anno_str = annotation.split(":")
    anno_dict = {str(i + 1): x for i, x in enumerate(anno_str.split(";"))}
    annotate(data, anno_name, clust_name, anno_dict)
    write_output(data, input_file, whitelist = ["obs"])



import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import time

from sccloud.io import read_input


def convert_to_parquet(data, output_name, nthreads):
    data.obs.index.name = "barcode"
    df = data.obs.reset_index()
    if "X_pca" in data.obsm.keys():
        df["PCA_X"] = data.obsm["X_pca"][:, 0]
        df["PCA_Y"] = data.obsm["X_pca"][:, 1]
    if "X_rpca" in data.obsm.keys():
        df["RPCA_X"] = data.obsm["X_rpca"][:, 0]
        df["RPCA_Y"] = data.obsm["X_rpca"][:, 1]
    if "X_tsne" in data.obsm.keys():
        df["TSNE_X"] = data.obsm["X_tsne"][:, 0]
        df["TSNE_Y"] = data.obsm["X_tsne"][:, 1]
    if "X_fitsne" in data.obsm.keys():
        df["FITSNE_X"] = data.obsm["X_fitsne"][:, 0]
        df["FITSNE_Y"] = data.obsm["X_fitsne"][:, 1]
    if "X_umap" in data.obsm.keys():
        df["UMAP_X"] = data.obsm["X_umap"][:, 0]
        df["UMAP_Y"] = data.obsm["X_umap"][:, 1]
    if "X_fle" in data.obsm.keys():
        df["FLE_X"] = data.obsm["X_fle"][:, 0]
        df["FLE_Y"] = data.obsm["X_fle"][:, 1]
    if "X_diffmap_pca" in data.obsm.keys():
        df["DIFFMAP_X"] = data.obsm["X_diffmap_pca"][:, 0]
        df["DIFFMAP_Y"] = data.obsm["X_diffmap_pca"][:, 1]
        df["DIFFMAP_Z"] = data.obsm["X_diffmap_pca"][:, 2]
    metadata_table = pa.Table.from_pandas(df, nthreads=nthreads)

    df_expr = pd.DataFrame(data=data.X.toarray(), columns=data.var_names)
    parquet_table = pa.Table.from_pandas(df_expr, nthreads=nthreads)

    for i in range(metadata_table.num_columns - 1, 0, -1):
        column = metadata_table[i]
        if column.name != "__index_level_0__":
            parquet_table = parquet_table.add_column(0, column)

    output_file = output_name + ".parquet"
    pq.write_table(parquet_table, output_file)
    print(output_file + " is written!")


def run_conversion(input_h5ad_file, output_name, nthreads):
    start = time.time()
    data = read_input(input_h5ad_file)
    end = time.time()
    print(
        "Time spent for loading the expression matrix is {:.2f}s.".format(end - start)
    )

    start = time.time()
    convert_to_parquet(data, output_name, nthreads)
    end = time.time()
    print("Time spent on generating the PARQUET file is {:.2f}s.".format(end - start))
